#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Settings window holding all settings tab"""
import gettext
import gi
import logging
import pkg_resources
import sys
import os
import json
from .autostart import Autostart, BazziteAutostart
from .draggable_window import DraggableWindow
from .draggable_window_wayland import DraggableWindowWayland

from configparser import ConfigParser
gi.require_version("Gtk", "3.0")
# pylint: disable=wrong-import-position,wrong-import-order
from gi.repository import Gtk, Gdk, Gio  # nopep8

log = logging.getLogger(__name__)
t = gettext.translation(
    'default', pkg_resources.resource_filename('discover_overlay', 'locales'), fallback=True)
_ = t.gettext


class MainSettingsWindow():
    """Settings class"""

    def __init__(self, config_file, rpc_file, channel_file, args):
        self.args = args
        # Detect Bazzite autostart
        self.alternative_autostart = os.path.exists(
            "/etc/default/discover-overlay")
        # Detect flatpak en
        self.disable_autostart = 'container' in os.environ
        self.icon_name = "discover-overlay"
        self.tray_icon_name = "discover-overlay-tray"

        self.spinning_focus = None
        self.scale_focus = None

        icon_theme = Gtk.IconTheme.get_default()
        icon_theme.add_resource_path(os.path.expanduser(
            '~/.local/share/pipx/venvs/discover-overlay/share/icons'))
        if not icon_theme.has_icon("discover-overlay"):
            log.error("No icon found in theme")
            self.icon_name = 'user-info'
        if not icon_theme.has_icon(self.tray_icon_name):
            log.error("No tray icon found in theme")
            self.tray_icon_name = 'user-info'
        self.steamos = False
        self.voice_placement_window = None
        self.text_placement_window = None
        self.tray = None  # Systemtray as fallback
        self.ind = None  # AppIndicator
        if self.alternative_autostart:
            self.autostart_helper = BazziteAutostart()
        else:
            self.autostart_helper = Autostart("discover_overlay")

        self.autostart_helper_conf = Autostart("discover_overlay_configure")
        self.ind = None
        self.guild_ids = []
        self.channel_ids = []
        self.current_guild = "0"
        self.current_channel = "0"
        self.hidden_overlay = False

        self.menu = self.make_menu()
        self.make_sys_tray_icon(self.menu)

        self.config_file = config_file
        self.rpc_file = rpc_file
        self.channel_file = channel_file

        self.loading_config = False

        builder = Gtk.Builder.new_from_file(pkg_resources.resource_filename(
            'discover_overlay', 'glade/settings.glade'))
        window = builder.get_object("settings_window")
        window.connect("destroy", self.close_window)
        window.connect("delete-event", self.close_window)

        window.set_default_size(1280, 800)

        # Make an array of all named widgets
        self.widget = {}
        for widget in builder.get_objects():
            if widget.find_property("name"):
                name = widget.get_property("name")
                if name == "":
                    log.error("Unnamed widget. All widgets must be named")
                    log.info(widget)
                self.widget[name] = widget

                # Translate labels and buttons
                if name.endswith("_label"):
                    widget.set_label(_(widget.get_label()))
                if name.endswith("_button"):
                    widget.set_label(_(widget.get_label()))
                if name.endswith("_all"):
                    widget.set_label(_(widget.get_label()))

        self.widget['overview_main_text'].set_markup("<span size=\"larger\">%s (%s)</span>\n\n%s\n\n%s (<a href=\"https://discord.gg/jRKWMuDy5V\">https://discord.gg/jRKWMuDy5V</a>) %s (<a href=\"https://github.com/trigg/Discover\">https://github.com/trigg/Discover</a>)\n\n\n\n\n\n" % (
            _("Welcome to Discover Overlay"),
            pkg_resources.get_distribution('discover_overlay').version,
            _("Discover-Overlay is a GTK3 overlay written in Python3. It can be configured to show who is currently talking on discord or it can be set to display text and images from a preconfigured channel. It is fully customisable and can be configured to display anywhere on the screen. We fully support X11 and wlroots based environments. We felt the need to make this project due to the shortcomings in support on Linux by the official discord client."),
            _("Please visit our discord"),
            _(" for support. Or open an issue on our GitHub ")
        ))

        screen = window.get_screen()
        screen_type = "%s" % (screen)
        self.is_wayland = False
        if "Wayland" in screen_type:
            self.is_wayland = True
        self.window = window

        if "GAMESCOPE_WAYLAND_DISPLAY" in os.environ:
            self.steamos = True
            log.info(
                "GameScope session detected. Enabling steam and gamescope integration")
            self.steamos = True
            settings = Gtk.Settings.get_default()
            if settings:
                settings.set_property(
                    "gtk-application-prefer-dark-theme", Gtk.true)
            self.set_steamos_window_size()

            # Larger fonts needed
            css = Gtk.CssProvider.new()
            css.load_from_data(bytes("* { font-size:18px; }", "utf-8"))
            window.get_style_context().add_provider(
                css, Gtk.STYLE_PROVIDER_PRIORITY_USER)
        else:
            self.widget['overview_close_button'].hide()

        self.super_focus = Gtk.CssProvider.new()
        self.super_focus.load_from_data(
            bytes(
                """scale { background-color: rgba(100%, 0%, 0%, 0.3); background-image:unset; }
                   spinbutton { background-color: rgba(100%, 0%, 0%, 0.3); background-image:unset;}
                """, "utf-8"))

        # Fill monitor & guild menus
        self.populate_monitor_menus()
        window.get_screen().connect("monitors-changed", self.populate_monitor_menus)

        channel_file = Gio.File.new_for_path(channel_file)
        self.monitor_channel = channel_file.monitor_file(0, None)
        self.monitor_channel.connect("changed", self.populate_guild_menu)

        self.server_handler = self.widget['text_server'].connect(
            'changed', self.text_server_changed)
        self.channel_handler = self.widget['text_channel'].connect(
            'changed', self.text_channel_changed)
        self.hidden_overlay_handler = self.widget['core_hide_overlay'].connect(
            'toggled', self.core_hide_overlay_changed)

        self.read_config()

        self.populate_guild_menu()

        builder.connect_signals(self)
        window.connect('key-press-event', self.keypress_in_settings)

        if '--minimized' in self.args:
            self.start_minimized = True
        if not self.start_minimized or not self.show_sys_tray_icon:
            window.show()

        if self.icon_name != 'discover-overlay':
            self.widget['overview_image'].set_from_icon_name(
                self.icon_name, Gtk.IconSize.DIALOG)
            self.widget['window'].set_default_icon_name(self.icon_name)

    def set_steamos_window_size(self):
        # Huge bunch of assumptions.
        # Gamescope only has one monitor
        # Gamescope has no scale factor
        display = Gdk.Display.get_default()
        if "get_monitor" in dir(display):
            monitor = display.get_monitor(0)
            if monitor:
                geometry = monitor.get_geometry()
                scale_factor = monitor.get_scale_factor()
                log.info("%d %d" % (geometry.width, geometry.height))
                self.window.set_size_request(geometry.width, geometry.height)

    def keypress_in_settings(self, window, event):
        if self.spinning_focus:
            match event.keyval:
                case Gdk.KEY_Right:
                    step = self.spinning_focus.get_increments().step
                    value = self.spinning_focus.get_value()
                    self.spinning_focus.set_value(value + step)
                    pass
                case Gdk.KEY_Left:
                    step = self.spinning_focus.get_increments().step
                    value = self.spinning_focus.get_value()
                    self.spinning_focus.set_value(value - step)
                    pass
                case Gdk.KEY_Up:
                    step = self.spinning_focus.get_increments().step
                    value = self.spinning_focus.get_value()
                    self.spinning_focus.set_value(value + step)
                case Gdk.KEY_Down:
                    step = self.spinning_focus.get_increments().step
                    value = self.spinning_focus.get_value()
                    self.spinning_focus.set_value(value - step)
                case Gdk.KEY_space:
                    self.spinning_focus.get_style_context().remove_provider(self.super_focus)
                    self.spinning_focus = None
                case Gdk.KEY_Escape:
                    self.spinning_focus.get_style_context().remove_provider(self.super_focus)

                    self.spinning_focus = None
        elif self.scale_focus:
            match event.keyval:
                case Gdk.KEY_Right:
                    value = self.scale_focus.get_value()
                    self.scale_focus.set_value(value + 0.1)
                    pass
                case Gdk.KEY_Left:
                    value = self.scale_focus.get_value()
                    self.scale_focus.set_value(value - 0.1)
                    pass
                case Gdk.KEY_Up:
                    value = self.scale_focus.get_value()
                    self.scale_focus.set_value(value + 0.1)
                case Gdk.KEY_Down:
                    value = self.scale_focus.get_value()
                    self.scale_focus.set_value(value - 0.1)
                case Gdk.KEY_space:
                    self.scale_focus.get_style_context().remove_provider(self.super_focus)
                    self.scale_focus = None
                case Gdk.KEY_Escape:
                    self.scale_focus.get_style_context().remove_provider(self.super_focus)
                    self.scale_focus = None
        else:
            match event.keyval:
                case Gdk.KEY_Left:
                    window.do_move_focus(window, Gtk.DirectionType.LEFT)
                case Gdk.KEY_Right:
                    window.do_move_focus(window, Gtk.DirectionType.RIGHT)
                case Gdk.KEY_Up:
                    window.do_move_focus(window, Gtk.DirectionType.UP)
                case Gdk.KEY_Down:
                    window.do_move_focus(window, Gtk.DirectionType.DOWN)
                case Gdk.KEY_F1:
                    self.widget['notebook'].prev_page()
                case Gdk.KEY_F2:
                    self.widget['notebook'].next_page()
                case Gdk.KEY_Escape:
                    return True
                case Gdk.KEY_space:
                    widget = self.window.get_focus()
                    if widget:
                        # I really want there to be a better way...
                        widget_type = "%s" % (widget)
                        if 'Gtk.SpinButton' in widget_type:
                            self.spinning_focus = widget

                            widget.get_style_context().add_provider(
                                self.super_focus, Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION)
                            return True
                        elif 'Gtk.Scale' in widget_type:
                            self.scale_focus = widget
                            widget.get_style_context().add_provider(
                                self.super_focus, Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION)
                            return True
                    return False
                case _:
                    return False
        return True

    def request_channels_from_guild(self, guild_id):
        with open(self.rpc_file, 'w') as f:
            f.write('--rpc --guild-request=%s' % (guild_id))

    def populate_guild_menu(self, _a=None, _b=None, _c=None, _d=None):
        g = self.widget['text_server']
        c = self.widget['text_channel']
        g.handler_block(self.server_handler)
        c.handler_block(self.channel_handler)
        try:
            with open(self.channel_file, "r") as tfile:
                data = tfile.readlines()
                if len(data) >= 1:
                    data = json.loads(data[0])
                    self.guild_ids = []
                    self.channel_ids = []
                    g.remove_all()
                    c.remove_all()
                    for guild in data['guild'].values():
                        g.append_text(guild['name'])
                        self.guild_ids.append(guild['id'])
                        if guild['id'] == self.current_guild and 'channels' in guild:
                            for channel in guild['channels']:
                                c.append_text(channel['name'])
                                self.channel_ids.append(channel['id'])
        except FileNotFoundError:
            pass

        if self.current_guild != "0" and self.current_guild in self.guild_ids:
            g.set_active(self.guild_ids.index(self.current_guild))

        if self.current_channel != "0" and self.current_channel in self.channel_ids:
            c.set_active(self.channel_ids.index(self.current_channel))

        g.handler_unblock(self.server_handler)
        c.handler_unblock(self.channel_handler)

    def populate_monitor_menus(self, _a=None, _b=None):
        v = self.widget['voice_monitor']
        t = self.widget['text_monitor']
        m = self.widget['notification_monitor']

        v_value = v.get_active()
        t_value = t.get_active()
        m_value = m.get_active()

        v.remove_all()
        t.remove_all()
        m.remove_all()

        v.append_text("Any")
        t.append_text("Any")
        m.append_text("Any")

        display = Gdk.Display.get_default()
        if "get_n_monitors" in dir(display):
            count_monitors = display.get_n_monitors()
            if count_monitors >= 1:
                for i in range(0, count_monitors):
                    v.append_text(display.get_monitor(i).get_model())
                    t.append_text(display.get_monitor(i).get_model())
                    m.append_text(display.get_monitor(i).get_model())

        v.set_active(v_value)
        t.set_active(t_value)
        m.set_active(m_value)

    def close_window(self, widget=None, event=None):
        """
        Hide the settings window for use at a later date
        """
        self.window.hide()
        if self.ind == None and self.tray == None:
            sys.exit(0)
        if self.ind != None:
            # pylint: disable=import-outside-toplevel
            from gi.repository import AppIndicator3
            if self.ind.get_status() == AppIndicator3.IndicatorStatus.PASSIVE:
                sys.exit(0)
        return True

    def close_app(self, widget=None, event=None):
        sys.exit(0)

    def present_settings(self, _a=None):
        """
        Show the settings window
        """
        self.widget['notebook'].set_current_page(0)
        self.window.show()

    def set_alignment_labels(self, horz):
        m1 = self.widget['voice_align_1'].get_model()
        m2 = self.widget['voice_align_2'].get_model()
        i = m1.get_iter_first()
        i2 = m2.get_iter_first()
        if horz:
            m1.set_value(i, 0, _("Top"))
            i = m1.iter_next(i)
            m1.set_value(i, 0, _("Bottom"))

            m2.set_value(i2, 0, _("Left"))
            i2 = m2.iter_next(i2)
            m2.set_value(i2, 0, _("Middle"))
            i2 = m2.iter_next(i2)
            m2.set_value(i2, 0, _("Right"))
        else:
            m1.set_value(i, 0, _("Left"))
            i = m1.iter_next(i)
            m1.set_value(i, 0, _("Right"))

            m2.set_value(i2, 0, _("Top"))
            i2 = m2.iter_next(i2)
            m2.set_value(i2, 0, _("Middle"))
            i2 = m2.iter_next(i2)
            m2.set_value(i2, 0, _("Bottom"))

    def read_config(self):
        self.loading_config = True

        # Read config and put into gui
        config = ConfigParser(interpolation=None)
        config.read(self.config_file)

        # Read Voice section

        self.voice_floating_x = config.getfloat(
            "main", "floating_x", fallback=0)
        self.voice_floating_y = config.getfloat(
            "main", "floating_y", fallback=0)
        self.voice_floating_w = config.getfloat(
            "main", "floating_w", fallback=400)
        self.voice_floating_h = config.getfloat(
            "main", "floating_h", fallback=400)

        self.widget['voice_anchor_float'].set_active(
            0 if config.getboolean("main", "floating", fallback=False) else 1)
        self.update_floating_ahchor()

        self.widget['voice_align_1'].set_active(
            config.getboolean("main", "rightalign", fallback=False))
        self.widget['voice_align_2'].set_active(
            config.getint("main", "topalign", fallback=1))

        monitor = 0
        try:
            config.getint("main", "monitor", fallback=0)
        except:
            pass

        self.widget['voice_monitor'].set_active(monitor+1)

        font = config.get("main", "font", fallback=None)
        if font:
            self.widget['voice_font'].set_font(font)
        title_font = config.get("main", "title_font", fallback=None)
        if title_font:
            self.widget['voice_title_font'].set_font(title_font)

        self.widget['voice_icon_spacing'].set_value(
            config.getint("main", "icon_spacing", fallback=8))

        self.widget['voice_text_padding'].set_value(
            config.getint("main", "text_padding", fallback=6))

        self.widget['voice_text_vertical_offset'].set_value(
            config.getint("main", "text_baseline_adj", fallback=0))

        self.widget['voice_vertical_padding'].set_value(
            config.getint("main", "vert_edge_padding", fallback=0))

        self.widget['voice_horizontal_padding'].set_value(
            config.getint("main", "horz_edge_padding", fallback=0))

        horz = config.getboolean("main", "horizontal", fallback=False)
        self.set_alignment_labels(horz)
        self.widget['voice_display_horizontally'].set_active(horz)

        self.widget['voice_highlight_self'].set_active(
            config.getboolean("main", "highlight_self", fallback=False))

        self.widget['voice_display_speakers_only'].set_active(
            config.getboolean("main", "only_speaking", fallback=False))

        self.widget['voice_display_speakers_grace_period'].set_value(
            config.getint("main", "only_speaking_grace", fallback=0))

        self.widget['voice_show_test_content'].set_active(
            config.getboolean("main", "show_dummy", fallback=False))

        self.widget['voice_talking_foreground'].set_rgba(self.make_colour(config.get(
            "main", "fg_hi_col", fallback="[1.0,1.0,1.0,1.0]")))

        self.widget['voice_talking_background'].set_rgba(self.make_colour(config.get(
            "main", "hi_col", fallback="[0.0,0.0,0.0,0.5]")))

        self.widget['voice_talking_border'].set_rgba(self.make_colour(config.get(
            "main", "tk_col", fallback="[0.0,0.7,0.0,1.0]")))

        self.widget['voice_idle_foreground'].set_rgba(self.make_colour(config.get(
            "main", "fg_col", fallback="[1.0,1.0,1.0,1.0]")))

        self.widget['voice_idle_background'].set_rgba(self.make_colour(config.get(
            "main", "bg_col", fallback="[0.0,0.0,0.0,0.5]")))

        self.widget['voice_idle_border'].set_rgba(self.make_colour(config.get(
            "main", "bo_col", fallback="[0.0,0.0,0.0,0.0]")))

        self.widget['voice_mute_foreground'].set_rgba(self.make_colour(config.get(
            "main", "mt_col", fallback="[0.6,0.0,0.0,1.0]")))

        self.widget['voice_mute_background'].set_rgba(self.make_colour(config.get(
            "main", "mt_bg_col", fallback="[0.0,0.0,0.0,0.5]")))

        self.widget['voice_avatar_background'].set_rgba(self.make_colour(config.get(
            "main", "avatar_bg_col", fallback="[0.0,0.0,0.0,0.0]")))

        self.widget['voice_avatar_opacity'].set_value(config.getfloat(
            "main", "icon_transparency", fallback=1.0))

        self.widget['voice_nick_length'].set_value(
            config.getint("main", "nick_length", fallback=32))

        self.widget['voice_avatar_size'].set_value(
            config.getint("main", "avatar_size", fallback=48))

        show_name = not config.getboolean("main", "icon_only", fallback=False)
        self.widget['voice_display_icon_only'].set_active(show_name)
        self.voice_show_name_hide_others(show_name)

        self.widget['voice_square_avatar'].set_active(config.getboolean(
            "main", "square_avatar", fallback=True))

        self.widget['voice_fancy_avatar_shapes'].set_active(config.getboolean("main",
                                                                              "fancy_border", fallback=True))

        self.widget['voice_order_avatars_by'].set_active(
            config.getint("main", "order", fallback=0))

        self.widget['voice_border_width'].set_value(
            config.getint("main", "border_width", fallback=2))

        self.widget['voice_overflow_style'].set_active(
            config.getint("main", "overflow", fallback=0))

        self.widget['voice_show_title'].set_active(config.getboolean(
            "main", "show_title", fallback=False))

        show_avatar = config.getboolean(
            "main", "show_avatar", fallback=True)
        self.widget['voice_show_avatar'].set_active(show_avatar)
        self.voice_show_avatar_hide_others(show_avatar)

        self.widget['voice_show_connection_status'].set_active(config.getboolean(
            "main", "show_connection", fallback=False))

        self.widget['voice_show_disconnected'].set_active(config.getboolean(
            "main", "show_disconnected", fallback=False))

        self.widget['voice_dummy_count'].set_value(
            config.getint("main", "dummy_count", fallback=50))

        self.widget['voice_inactive_fade'].set_active(
            config.getboolean("main", "fade_out_inactive", fallback=False)
        )
        self.widget['voice_inactive_opacity'].set_value(
            config.getfloat("main", "fade_out_limit", fallback=0.3)
        )
        self.widget['voice_inactive_time'].set_value(
            config.getint("main", "inactive_time", fallback=10)
        )
        self.widget['voice_inactive_fade_time'].set_value(
            config.getint("main", "inactive_fade_time", fallback=30)
        )
        self.widget['voice_hide_mouseover'].set_active(
            config.getboolean("main", "autohide", fallback=False)
        )
        self.widget['voice_show_mouseover'].set_value(
            config.getint("main", "autohide_timer", fallback=5)
        )

        # Read Text section

        self.text_floating_x = config.getfloat(
            "text", "floating_x", fallback=0)
        self.text_floating_y = config.getfloat(
            "text", "floating_y", fallback=0)
        self.text_floating_w = config.getfloat(
            "text", "floating_w", fallback=400)
        self.text_floating_h = config.getfloat(
            "text", "floating_h", fallback=400)

        self.widget['text_enable'].set_active(
            config.getboolean("text", "enabled", fallback=False))

        self.widget['text_popup_style'].set_active(
            config.getboolean("text", "popup_style", fallback=False))

        self.widget['text_popup_time'].set_value(
            config.getint("text", "text_time", fallback=30)
        )

        self.current_guild = config.get("text", "guild", fallback="0")

        self.current_channel = config.get("text", "channel", fallback="0")

        font = config.get("text", "font", fallback=None)
        if font:
            self.widget['text_font'].set_font(font)

        self.widget['text_colour'].set_rgba(self.make_colour(config.get(
            "text", "fg_col", fallback="[1.0,1.0,1.0,1.0]")))

        self.widget['text_background_colour'].set_rgba(self.make_colour(config.get(
            "text", "bg_col", fallback="[0.0,0.0,0.0,0.5]")))

        monitor = 0
        try:
            config.getint("text", "monitor", fallback=0)
        except:
            pass

        self.widget['text_monitor'].set_active(monitor+1)

        self.widget['text_show_attachments'].set_active(config.getboolean(
            "text", "show_attach", fallback=True))

        self.widget['text_line_limit'].set_value(
            config.getint("text", "line_limit", fallback=20))

        self.widget['text_hide_mouseover'].set_active(
            config.getboolean("text", "autohide", fallback=False)
        )
        self.widget['text_show_mouseover'].set_value(
            config.getint("text", "autohide_timer", fallback=5)
        )

        # Read Notification section
        self.widget['notification_enable'].set_active(
            config.getboolean("notification", "enabled", fallback=False))

        self.widget['notification_reverse_order'].set_active(
            config.getboolean("notification", "rev", fallback=False))

        self.widget['notification_popup_timer'].set_value(
            config.getint("notification", "text_time", fallback=10))

        self.widget['notification_limit_popup_width'].set_value(
            config.getint("notification", "limit_width", fallback=400))

        font = config.get("notification", "font", fallback=None)
        if font:
            self.widget['notification_font'].set_font(font)

        self.widget['notification_text_colour'].set_rgba(self.make_colour(config.get(
            "notification", "fg_col", fallback="[1.0,1.0,1.0,1.0]")))
        self.widget['notification_background_colour'].set_rgba(self.make_colour(config.get(
            "notification", "bg_col", fallback="[0.0,0.0,0.0,0.5]")))

        monitor = 0
        try:
            config.getint("notification", "monitor", fallback=0)
        except:
            pass

        self.widget['notification_monitor'].set_active(monitor+1)

        self.widget['notification_align_1'].set_active(config.getboolean(
            "notification", "rightalign", fallback=True))

        self.widget['notification_align_2'].set_active(
            config.getint("notification", "topalign", fallback=2))

        self.widget['notification_show_icon'].set_active(
            config.getboolean("notification", "show_icon", fallback=True))

        self.widget['notification_icon_position'].set_active(config.getboolean(
            "notification", "icon_left", fallback=True))

        self.widget['notification_icon_padding'].set_value(config.getint(
            "notification", "icon_padding", fallback=8))

        self.widget['notification_icon_size'].set_value(config.getint(
            "notification", "icon_size", fallback=32))

        self.widget['notification_padding_between'].set_value(config.getint(
            "notification", "padding", fallback=8))

        self.widget['notification_border_radius'].set_value(config.getint(
            "notification", "border_radius", fallback=8))

        self.widget['notification_show_test_content'].set_active(config.getboolean(
            "notification", "show_dummy", fallback=False))

        # Read Core section

        self.widget['core_run_on_startup'].set_active(
            self.autostart_helper.is_auto())

        self.widget['core_run_conf_on_startup'].set_active(
            self.autostart_helper_conf.is_auto())

        if self.disable_autostart:
            self.widget['core_run_on_startup'].set_sensitive(False)
            self.widget['core_run_conf_on_startup'].set_sensitive(False)

        self.widget['core_force_xshape'].set_active(
            config.getboolean("general", "xshape", fallback=False))

        self.show_sys_tray_icon = config.getboolean(
            "general", "showsystray", fallback=True)
        self.set_sys_tray_icon_visible(self.show_sys_tray_icon)
        self.widget['core_show_tray_icon'].set_active(self.show_sys_tray_icon)

        self.hidden_overlay = config.getboolean(
            "general", "hideoverlay", fallback=False)
        self.update_toggle_overlay()

        self.start_minimized = config.getboolean(
            "general", "start_min", fallback=False)

        self.widget['core_settings_min'].set_active(self.start_minimized)

        self.widget['core_settings_min'].set_sensitive(self.show_sys_tray_icon)

        if 'XDG_SESSION_DESKTOP' in os.environ and os.environ['XDG_SESSION_DESKTOP'] == 'cinnamon':
            self.widget['voice_anchor_float'].set_sensitive(False)

        self.widget['core_audio_assist'].set_active(
            config.getboolean("general", "audio_assist", fallback=False))

        self.loading_config = False

    def make_colour(self, col):
        col = json.loads(col)
        return Gdk.RGBA(col[0], col[1], col[2], col[3])

    def parse_guild_ids(self, guild_ids_str):
        """Parse the guild_ids from a str and return them in a list"""
        guild_ids = []
        for guild_id in guild_ids_str.split(","):
            guild_id = guild_id.strip()
            if guild_id != "":
                guild_ids.append(guild_id)
        return guild_ids

    def get_monitor_obj(self, idx):
        """
        Helper function to find the monitor object of the monitor
        """
        display = Gdk.Display.get_default()
        return display.get_monitor(idx)

    def make_sys_tray_icon(self, menu):
        """
        Attempt to create an AppIndicator icon, failing that attempt to make
        a systemtray icon
        """
        try:
            gi.require_version('AppIndicator3', '0.1')
            # pylint: disable=import-outside-toplevel
            from gi.repository import AppIndicator3
            self.ind = AppIndicator3.Indicator.new(
                "discover_overlay",
                self.tray_icon_name,
                AppIndicator3.IndicatorCategory.APPLICATION_STATUS)
            self.ind.set_title(_("Discover Overlay Configuration"))
            # Hide for now since we don't know if it should be shown yet
            self.ind.set_status(AppIndicator3.IndicatorStatus.PASSIVE)
            self.ind.set_menu(menu)
        except (ImportError, ValueError) as exception:
            # Create System Tray
            log.info("Falling back to Systray : %s", exception)
            self.tray = Gtk.StatusIcon.new_from_icon_name(
                self.tray_icon_name)
            self.tray.connect('popup-menu', self.show_menu)
            self.tray.set_title(_("Discover Overlay Configuration"))
            # Hide for now since we don't know if it should be shown yet
            self.tray.set_visible(False)

    def show_menu(self, obj, button, time):
        """
        Show menu when System Tray icon is clicked
        """
        self.menu.show_all()
        self.menu.popup(
            None, None, Gtk.StatusIcon.position_menu, obj, button, time)

    def set_sys_tray_icon_visible(self, visible):
        """
        Sets whether the tray icon is visible
        """
        if self.ind is not None:
            # pylint: disable=import-outside-toplevel
            from gi.repository import AppIndicator3
            self.ind.set_status(
                AppIndicator3.IndicatorStatus.ACTIVE if visible else AppIndicator3.IndicatorStatus.PASSIVE)
        elif self.tray is not None:
            self.tray.set_visible(visible)

    def make_menu(self):
        """
        Create System Menu
        """
        menu = Gtk.Menu()
        settings_opt = Gtk.MenuItem.new_with_label(_("Settings"))
        self.toggle_opt = Gtk.MenuItem.new_with_label(_("Hide overlay"))
        close_overlay_opt = Gtk.MenuItem.new_with_label(_("Close Overlay"))
        close_opt = Gtk.MenuItem.new_with_label(_("Close Settings"))

        menu.append(settings_opt)
        menu.append(self.toggle_opt)
        menu.append(close_overlay_opt)
        menu.append(close_opt)

        settings_opt.connect("activate", self.present_settings)
        self.toggle_opt.connect("activate", self.toggle_overlay)
        close_overlay_opt.connect("activate", self.close_overlay)
        close_opt.connect("activate", self.close_app)
        menu.show_all()
        return menu

    def toggle_overlay(self, _a=None, _b=None):
        self.hidden_overlay = not self.hidden_overlay
        self.config_set("general", "hideoverlay", "%s" % (self.hidden_overlay))
        self.update_toggle_overlay()

    def update_toggle_overlay(self, _a=None, _b=None):
        self.widget['core_hide_overlay'].handler_block(
            self.hidden_overlay_handler)

        self.widget['core_hide_overlay'].set_active(self.hidden_overlay)

        self.widget['core_hide_overlay'].handler_unblock(
            self.hidden_overlay_handler)
        if self.hidden_overlay:
            self.toggle_opt.set_label(_("Show overlay"))
        else:
            self.toggle_opt.set_label(_("Hide overlay"))

    def close_overlay(self, _a=None, _b=None):
        with open(self.rpc_file, 'w') as f:
            f.write('--rpc --close')

    def voice_toggle_test_content(self, button):
        self.voice_overlay.set_show_dummy(button.get_active())
        self.show_dummy = button.get_active()
        if self.show_dummy:
            self.voice_overlay.set_enabled(True)
            self.voice_overlay.set_hidden(False)

    def overview_close(self, button):
        log.info("Quit pressed")
        self.close_overlay()

    def voice_place_window(self, button):
        if self.voice_placement_window:
            (pos_x, pos_y, width, height) = self.voice_placement_window.get_coords()
            self.voice_floating_x = pos_x
            self.voice_floating_y = pos_y
            self.voice_floating_w = width
            self.voice_floating_h = height

            config = ConfigParser(interpolation=None)
            config.read(self.config_file)
            if not "main" in config.sections():
                config.add_section("main")
            config.set("main", "floating_x", "%f" %
                       (self.voice_floating_x))
            config.set("main", "floating_y", "%f" %
                       (self.voice_floating_y))
            config.set("main", "floating_w", "%f" %
                       (self.voice_floating_w))
            config.set("main", "floating_h", "%f" %
                       (self.voice_floating_h))

            with open(self.config_file, 'w') as file:
                config.write(file)
            if button:
                button.set_label(_("Place Window"))
            self.voice_placement_window.close()
            self.voice_placement_window = None
            if self.steamos:
                self.window.show()
        else:
            if self.steamos:
                self.window.hide()
            if self.is_wayland or self.steamos:
                self.voice_placement_window = DraggableWindowWayland(
                    pos_x=self.voice_floating_x, pos_y=self.voice_floating_y,
                    width=self.voice_floating_w, height=self.voice_floating_h,
                    message=_("Place & resize this window then press Green!"), settings=self,
                    steamos=self.steamos,
                    monitor=self.widget['voice_monitor'].get_active()-1)
            else:
                self.voice_placement_window = DraggableWindow(
                    pos_x=self.voice_floating_x, pos_y=self.voice_floating_y,
                    width=self.voice_floating_w, height=self.voice_floating_h,
                    message=_("Place & resize this window then press Save!"), settings=self, monitor=self.widget['voice_monitor'].get_active()-1)
                if button:
                    button.set_label(_("Save this position"))

    def text_place_window(self, button):
        if self.text_placement_window:
            (pos_x, pos_y, width, height) = self.text_placement_window.get_coords()
            self.text_floating_x = pos_x
            self.text_floating_y = pos_y
            self.text_floating_w = width
            self.text_floating_h = height

            config = ConfigParser(interpolation=None)
            config.read(self.config_file)
            if not "text" in config.sections():
                config.add_section("text")
            config.set("text", "floating_x", "%f" %
                       (self.text_floating_x))
            config.set("text", "floating_y", "%f" %
                       (self.text_floating_y))
            config.set("text", "floating_w", "%f" %
                       (self.text_floating_w))
            config.set("text", "floating_h", "%f" %
                       (self.text_floating_h))

            with open(self.config_file, 'w') as file:
                config.write(file)
            if button:
                button.set_label(_("Place Window"))
            self.text_placement_window.close()
            self.text_placement_window = None
            if self.steamos:
                self.window.show()
        else:
            if self.steamos:
                self.window.hide()
            if self.is_wayland or self.steamos:
                self.text_placement_window = DraggableWindowWayland(
                    pos_x=self.text_floating_x, pos_y=self.text_floating_y,
                    width=self.text_floating_w, height=self.text_floating_h,
                    message=_("Place & resize this window then press Green!"), settings=self,
                    steamos=self.steamos,
                    monitor=self.widget['text_monitor'].get_active()-1)
            else:
                self.text_placement_window = DraggableWindow(
                    pos_x=self.text_floating_x, pos_y=self.text_floating_y,
                    width=self.text_floating_w, height=self.text_floating_h,
                    message=_("Place & resize this window then press Save!"), settings=self, monitor=self.widget['text_monitor'].get_active()-1)
                if button:
                    button.set_label(_("Save this position"))

    def change_placement(self, placement_window):
        if placement_window == self.text_placement_window:
            self.text_place_window(None)
        elif placement_window == self.voice_placement_window:
            self.voice_place_window(None)

    def text_server_refresh(self, button):
        with open(self.rpc_file, 'w') as f:
            f.write('--rpc --refresh-guilds')

    def config_set(self, context, key, value):
        if self.loading_config:
            return
        config = ConfigParser(interpolation=None)
        config.read(self.config_file)
        if not context in config.sections():
            config.add_section(context)
        config.set(context, key, value)
        with open(self.config_file, 'w') as file:
            config.write(file)

    def config_remove_section(self, context):
        if self.loading_config:
            return
        config = ConfigParser(interpolation=None)
        config.read(self.config_file)
        if context in config.sections():
            config.remove_section(context)
        else:
            log.error("Unable to remove section %s" % (context))
        with open(self.config_file, 'w') as file:
            config.write(file)

    def voice_anchor_float_changed(self, button):
        self.config_set("main", "floating", "%s" % (button.get_active() == 0))
        self.update_floating_ahchor()

    def update_floating_ahchor(self):
        floating = self.widget['voice_anchor_float'].get_active() == 0
        if 'XDG_SESSION_DESKTOP' in os.environ and os.environ['XDG_SESSION_DESKTOP'] == 'cinnamon':
            floating = True

        if floating:
            self.widget['voice_align_1'].hide()
            self.widget['voice_align_2'].hide()
            self.widget['voice_place_window_button'].show()
        else:
            self.widget['voice_align_1'].show()
            self.widget['voice_align_2'].show()
            self.widget['voice_place_window_button'].hide()

    def voice_monitor_changed(self, button):
        self.config_set("main", "monitor", "%s" % (button.get_active()-1))

    def voice_align_1_changed(self, button):
        self.config_set("main", "rightalign", "%s" % (button.get_active()))

    def voice_align_2_changed(self, button):
        self.config_set("main", "topalign", "%s" % (button.get_active()))

    def voice_font_changed(self, button):
        self.config_set("main", "font", button.get_font())

    def voice_title_font_changed(self, button):
        self.config_set("main", "title_font", button.get_font())

    def voice_icon_spacing_changed(self, button):
        self.config_set("main", "icon_spacing", "%s" %
                        (int(button.get_value())))

    def voice_text_padding_changed(self, button):
        self.config_set("main", "text_padding", "%s" %
                        (int(button.get_value())))

    def voice_text_vertical_offset_changed(self, button):
        self.config_set("main", "text_baseline_adj", "%s" %
                        (int(button.get_value())))

    def voice_vertical_padding_changed(self, button):
        self.config_set("main", "vert_edge_padding", "%s" %
                        (int(button.get_value())))

    def voice_horizontal_padding_changed(self, button):
        self.config_set("main", "horz_edge_padding", "%s" %
                        (int(button.get_value())))

    def voice_display_horizontally_changed(self, button):
        self.config_set("main", "horizontal", "%s" % (button.get_active()))
        self.set_alignment_labels(button.get_active())

    def voice_highlight_self_changed(self, button):
        self.config_set("main", "highlight_self", "%s" % (button.get_active()))

    def voice_display_speakers_only(self, button):
        self.config_set("main", "only_speaking", "%s" % (button.get_active()))

    def voice_display_speakers_grace_period(self, button):
        self.config_set("main", "only_speaking_grace", "%s" %
                        (int(button.get_value())))

    def voice_toggle_test_content(self, button):
        self.config_set("main", "show_dummy", "%s" % (button.get_active()))

    def voice_talking_foreground_changed(self, button):
        colour = button.get_rgba()
        colour = [colour.red, colour.green, colour.blue, colour.alpha]
        self.config_set("main", "fg_hi_col", json.dumps(colour))

    def voice_talking_background_changed(self, button):
        colour = button.get_rgba()
        colour = [colour.red, colour.green, colour.blue, colour.alpha]
        self.config_set("main", "hi_col", json.dumps(colour))

    def voice_talking_border_changed(self, button):
        colour = button.get_rgba()
        colour = [colour.red, colour.green, colour.blue, colour.alpha]
        self.config_set("main", "tk_col", json.dumps(colour))

    def voice_idle_foreground_changed(self, button):
        colour = button.get_rgba()
        colour = [colour.red, colour.green, colour.blue, colour.alpha]
        self.config_set("main", "fg_col", json.dumps(colour))

    def voice_idle_background_changed(self, button):
        colour = button.get_rgba()
        colour = [colour.red, colour.green, colour.blue, colour.alpha]
        self.config_set("main", "bg_col", json.dumps(colour))

    def voice_idle_border_changed(self, button):
        colour = button.get_rgba()
        colour = [colour.red, colour.green, colour.blue, colour.alpha]
        self.config_set("main", "bo_col", json.dumps(colour))

    def voice_mute_foreground_changed(self, button):
        colour = button.get_rgba()
        colour = [colour.red, colour.green, colour.blue, colour.alpha]
        self.config_set("main", "mt_col", json.dumps(colour))

    def voice_mute_background_changed(self, button):
        colour = button.get_rgba()
        colour = [colour.red, colour.green, colour.blue, colour.alpha]
        self.config_set("main", "mt_bg_col", json.dumps(colour))

    def voice_avatar_background_changed(self, button):
        colour = button.get_rgba()
        colour = [colour.red, colour.green, colour.blue, colour.alpha]
        self.config_set("main", "avatar_bg_col", json.dumps(colour))

    def voice_avatar_opacity_changed(self, button):
        self.config_set("main", "icon_transparency", "%.2f" %
                        (button.get_value()))

    def voice_avatar_size_changed(self, button):
        self.config_set("main", "avatar_size", "%s" %
                        (int(button.get_value())))

    def voice_nick_length_changed(self, button):
        self.config_set("main", "nick_length", "%s" %
                        (int(button.get_value())))

    def voice_display_icon_only_changed(self, button):
        self.config_set("main", "icon_only", "%s" % (not button.get_active()))
        self.voice_show_name_hide_others(button.get_active())

    def voice_square_avatar_changed(self, button):
        self.config_set("main", "square_avatar", "%s" % (button.get_active()))

    def voice_fancy_avatar_shapes_changed(self, button):
        self.config_set("main", "fancy_border", "%s" % (button.get_active()))

    def voice_order_avatars_by_changed(self, button):
        self.config_set("main", "order", "%s" % (button.get_active()))

    def voice_border_width_changed(self, button):
        self.config_set("main", "border_width", "%s" %
                        (int(button.get_value())))

    def voice_overflow_style_changed(self, button):
        self.config_set("main", "overflow", "%s" % (int(button.get_active())))

    def voice_show_title_changed(self, button):
        self.config_set("main", "show_title", "%s" % (button.get_active()))

    def voice_show_connection_status_changed(self, button):
        self.config_set("main", "show_connection", "%s" %
                        (button.get_active()))

    def voice_show_disconnected_changed(self, button):
        self.config_set("main", "show_disconnected", "%s" %
                        (button.get_active()))

    def voice_dummy_count_changed(self, button):
        self.config_set("main", "dummy_count", "%s" %
                        (int(button.get_value())))

    def voice_show_avatar_changed(self, button):
        self.config_set("main", "show_avatar", "%s" % (button.get_active()))
        self.voice_show_avatar_hide_others(button.get_active())

    def voice_show_name_hide_others(self, val):
        if val:
            # Show name options
            self.widget['voice_font'].set_sensitive(True)
            self.widget['voice_text_padding'].set_sensitive(True)
            self.widget['voice_text_vertical_offset'].set_sensitive(True)
            self.widget['voice_nick_length'].set_sensitive(True)
        else:
            # Hide name options
            self.widget['voice_font'].set_sensitive(False)
            self.widget['voice_text_padding'].set_sensitive(False)
            self.widget['voice_text_vertical_offset'].set_sensitive(False)
            self.widget['voice_nick_length'].set_sensitive(False)

    def voice_show_avatar_hide_others(self, val):
        if val:
            # Show avatar options
            self.widget['voice_square_avatar'].set_sensitive(True)
            self.widget['voice_fancy_avatar_shapes'].set_sensitive(True)
            self.widget['voice_avatar_size'].set_sensitive(True)
            self.widget['voice_avatar_opacity'].set_sensitive(True)
        else:
            # Hide avatar options
            self.widget['voice_square_avatar'].set_sensitive(False)
            self.widget['voice_fancy_avatar_shapes'].set_sensitive(False)
            self.widget['voice_avatar_size'].set_sensitive(False)
            self.widget['voice_avatar_opacity'].set_sensitive(False)

    def text_enable_changed(self, button):
        self.config_set("text", "enabled", "%s" % (button.get_active()))

    def text_popup_style_changed(self, button):
        self.config_set("text", "popup_style", "%s" % (button.get_active()))

    def text_popup_time_changed(self, button):
        self.config_set("text", "text_time", "%s" % (int(button.get_value())))

    def text_server_changed(self, button):
        if button.get_active() < 0:
            self.config_set("text", "guild", "0")
            return
        guild = self.guild_ids[button.get_active()]
        if guild and self.current_guild != guild:
            self.current_guild = guild
            self.config_set("text", "guild", guild)
            self.request_channels_from_guild(guild)

    def text_channel_changed(self, button):
        if button.get_active() < 0:
            self.config_set("text", "channel", "0")
            return
        channel = self.channel_ids[button.get_active()]
        if channel:
            self.current_channel = channel
            self.config_set("text", "channel", channel)

    def text_font_changed(self, button):
        self.config_set("text", "font", button.get_font())

    def text_colour_changed(self, button):
        colour = button.get_rgba()
        colour = [colour.red, colour.green, colour.blue, colour.alpha]
        self.config_set("text", "fg_col", json.dumps(colour))

    def text_background_colour_changed(self, button):
        colour = button.get_rgba()
        colour = [colour.red, colour.green, colour.blue, colour.alpha]
        self.config_set("text", "bg_col", json.dumps(colour))

    def text_monitor_changed(self, button):
        self.config_set("text", "monitor", "%s" % (button.get_active()-1))

    def text_show_attachments_changed(self, button):
        self.config_set("text", "show_attach", "%s" % (button.get_active()))

    def text_line_limit_changed(self, button):
        self.config_set("text", "line_limit", "%s" % (int(button.get_value())))

    def notification_enable_changed(self, button):
        self.config_set("notification", "enabled", "%s" %
                        (button.get_active()))

    def notification_reverse_order_changed(self, button):
        self.config_set("notification", "rev", "%s" % (button.get_active()))

    def notification_popup_timer_changed(self, button):
        self.config_set("notification", "text_time", "%s" %
                        (int(button.get_value())))

    def notification_limit_popup_width_changed(self, button):
        self.config_set("notification", "limit_width", "%s" %
                        (int(button.get_value())))

    def notification_font_changed(self, button):
        self.config_set("notification", "font", button.get_font())

    def notification_text_colour_changed(self, button):
        colour = button.get_rgba()
        colour = [colour.red, colour.green, colour.blue, colour.alpha]
        self.config_set("notification", "fg_col", json.dumps(colour))

    def notification_background_colour_changed(self, button):
        colour = button.get_rgba()
        colour = [colour.red, colour.green, colour.blue, colour.alpha]
        self.config_set("notification", "bg_col", json.dumps(colour))

    def notification_monitor_changed(self, button):
        self.config_set("notification", "monitor", "%s" %
                        (button.get_active()-1))

    def notification_align_1_changed(self, button):
        self.config_set("notification", "rightalign", "%s" %
                        (button.get_active()))

    def notification_align_2_changed(self, button):
        self.config_set("notification", "topalign", "%s" %
                        (button.get_active()))

    def notification_show_icon(self, button):
        self.config_set("notification", "show_icon", "%s" %
                        (button.get_active()))

    def notification_icon_position_changed(self, button):
        self.config_set("notification", "icon_left", "%s" %
                        (int(button.get_active() != 1)))

    def notification_icon_padding_changed(self, button):
        self.config_set("notification", "icon_padding", "%s" %
                        (int(button.get_value())))

    def notification_icon_size_changed(self, button):
        self.config_set("notification", "icon_size", "%s" %
                        (int(button.get_value())))

    def notification_padding_between_changed(self, button):
        self.config_set("notification", "padding", "%s" %
                        (int(button.get_value())))

    def notification_border_radius_changed(self, button):
        self.config_set("notification", "border_radius", "%s" %
                        (int(button.get_value())))

    def notification_show_test_content_changed(self, button):
        self.config_set("notification", "show_dummy", "%s" %
                        (button.get_active()))

    def core_run_on_startup_changed(self, button):
        self.autostart_helper.set_autostart(button.get_active())

    def core_run_conf_on_startup_changed(self, button):
        self.autostart_helper_conf.set_autostart(button.get_active())

    def core_force_xshape_changed(self, button):
        self.config_set("general", "xshape", "%s" % (button.get_active()))

    def core_show_tray_icon_changed(self, button):
        self.set_sys_tray_icon_visible(button.get_active())
        self.config_set("general", "showsystray", "%s" % (button.get_active()))
        self.widget['core_settings_min'].set_sensitive(button.get_active())

    def core_hide_overlay_changed(self, button):
        self.toggle_overlay()

    def core_settings_min_changed(self, button):
        self.config_set("general", "start_min", "%s" % (button.get_active()))

    def core_reset_all(self, button):
        self.config_remove_section("general")
        self.read_config()

    def voice_reset_all(self, button):
        self.config_remove_section("main")
        self.read_config()

    def text_reset_all(self, button):
        self.config_remove_section("text")
        self.read_config()

    def notification_reset_all(self, button):
        self.config_remove_section("notification")
        self.read_config()

    def voice_hide_mouseover_changed(self, button):
        self.config_set("main", "autohide", "%s" % (button.get_active()))

    def text_hide_mouseover_changed(self, button):
        self.config_set("text", "autohide", "%s" % (button.get_active()))

    def voice_mouseover_timeout_changed(self, button):
        self.config_set("main", "autohide_timer", "%s" %
                        (int(button.get_value())))

    def text_mouseover_timeout_changed(self, button):
        self.config_set("text", "autohide_timer", "%s" %
                        (int(button.get_value())))

    def inactive_fade_changed(self, button):
        self.config_set("main", "fade_out_inactive", "%s" %
                        (button.get_active()))

    def inactive_fade_opacity_changed(self, button):
        self.config_set("main", "fade_out_limit", "%.2f" %
                        (button.get_value()))

    def inactive_time_changed(self, button):
        self.config_set("main", "inactive_time", "%s" %
                        (int(button.get_value())))

    def inactive_fade_time_changed(self, button):
        self.config_set("main", "inactive_fade_time", "%s" %
                        (int(button.get_value())))

    def core_audio_assist_changed(self, button):
        self.config_set("general", "audio_assist", "%s" %
                        (button.get_active()))



#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
Overlay parent class. Helpful if we need more overlay
types without copy-and-pasting too much code
"""
import os
import sys
import logging
import gi
import cairo
import Xlib
from Xlib.display import Display
from Xlib import X, Xatom
gi.require_version("Gtk", "3.0")
# pylint: disable=wrong-import-position,wrong-import-order
from gi.repository import Gtk, Gdk, GLib  # nopep8
try:
    gi.require_version('GtkLayerShell', '0.1')
    from gi.repository import GtkLayerShell
except (ImportError, ValueError):
    pass

log = logging.getLogger(__name__)


class OverlayWindow(Gtk.Window):
    """
    Overlay parent class. Helpful if we need more overlay
    types without copy-and-pasting too much code
    """

    def detect_type(self):
        """
        Helper function to determine if Wayland is being used and return the Window type needed
        """
        window = Gtk.Window()
        screen = window.get_screen()
        screen_type = "%s" % (screen)
        self.is_wayland = False
        if "Wayland" in screen_type:
            self.is_wayland = True
            return Gtk.WindowType.TOPLEVEL
        return Gtk.WindowType.POPUP

    def __init__(self, discover, piggyback=None):
        Gtk.Window.__init__(self, type=self.detect_type())
        self.is_xatom_set = False

        self.discover = discover
        screen = self.get_screen()
        self.text_font = None
        self.text_size = None
        self.pos_x = None
        self.pos_y = None
        self.width = None
        self.height = None
        self.hidden = False
        self.enabled = False
        self.set_size_request(50, 50)
        self.hide_on_mouseover = True
        self.connect('draw', self.overlay_draw_pre)
        # Set RGBA
        screen = self.get_screen()
        visual = screen.get_rgba_visual()
        if not self.get_display().supports_input_shapes():
            log.info(
                "Input shapes not available. Quitting")
            sys.exit(1)
        if visual:
            # Set the visual even if we can't use it right now
            self.set_visual(visual)

        self.set_app_paintable(True)
        self.set_untouchable()
        self.set_skip_pager_hint(True)
        self.set_skip_taskbar_hint(True)
        self.set_keep_above(True)
        self.set_decorated(True)
        self.set_accept_focus(False)
        self.set_wayland_state()
        self.piggyback = None
        self.piggyback_parent = None
        if not piggyback:
            self.show_all()
            if discover.steamos:
                self.set_gamescope_xatom(1)
        self.monitor = 0
        self.align_right = True
        self.align_vert = 1
        self.floating = False
        self.force_xshape = False
        self.context = None

        self.redraw_id = None
        self.draw_blank = False
        self.timeout_mouse_over = 1

        self.timer_after_draw = None
        if piggyback:
            self.set_piggyback(piggyback)

        self.get_screen().connect("composited-changed", self.check_composite)
        self.get_screen().connect("monitors-changed", self.screen_changed)
        self.get_screen().connect("size-changed", self.screen_changed)
        if self.get_window():
            self.get_window().set_events(self.get_window().get_events()
                                         | Gdk.EventMask.ENTER_NOTIFY_MASK)
        self.connect("enter-notify-event", self.mouseover)
        self.connect("leave-notify-event", self.mouseout)
        self.mouse_over_timer = None

        # It shouldn't be possible, but let's not leave
        # this process hanging if it happens
        self.connect('destroy', self.window_exited)

    def window_exited(self, window=None):
        sys.exit(1)

    def set_gamescope_xatom(self, enabled):
        if self.piggyback_parent:
            return

        if enabled == self.is_xatom_set:
            return
        self.is_xatom_set = enabled
        display = Display()
        atom = display.intern_atom("GAMESCOPE_EXTERNAL_OVERLAY")
        opaq = display.intern_atom("_NET_WM_WINDOW_OPACITY")

        if self.get_toplevel().get_window():
            topw = display.create_resource_object(
                "window", self.get_toplevel().get_window().get_xid())

            topw.change_property(atom,
                                 Xatom.CARDINAL, 32,
                                 [enabled], X.PropModeReplace)
            log.info("Setting GAMESCOPE_EXTERNAL_OVERLAY to %s", enabled)
            display.sync()
        else:
            log.warn("Unable to set GAMESCOPE_EXTERNAL_OVERLAY")

    def set_wayland_state(self):
        """
        If wayland is in use then attempt to set up a GtkLayerShell
        """
        if self.is_wayland:
            if not GtkLayerShell.is_supported():
                log.info(
                    "GTK Layer Shell is not supported on this wayland compositor")
                log.info("Currently not possible: Gnome, Weston")
                sys.exit(0)
            if not GtkLayerShell.is_layer_window(self):
                GtkLayerShell.init_for_window(self)
            GtkLayerShell.set_layer(self, GtkLayerShell.Layer.OVERLAY)
            GtkLayerShell.set_anchor(self, GtkLayerShell.Edge.LEFT, True)
            GtkLayerShell.set_anchor(self, GtkLayerShell.Edge.RIGHT, True)
            GtkLayerShell.set_anchor(self, GtkLayerShell.Edge.BOTTOM, True)
            GtkLayerShell.set_anchor(self, GtkLayerShell.Edge.TOP, True)

    def set_piggyback(self, other_overlay):
        other_overlay.piggyback = self
        self.piggyback_parent = other_overlay

    def has_content(self):
        return False

    def overlay_draw_pre(self, _w, context, data=None):
        content = self.has_content()
        if self.piggyback and self.piggyback.has_content():
            content = True
        if self.discover.steamos:
            if not content:
                self.set_gamescope_xatom(0)
            else:
                if not self.hidden and self.enabled:
                    self.set_gamescope_xatom(1)
        # If we're hiding on mouseover, allow mouse-in
        if self.hide_on_mouseover:
            # We've mouse-overed
            if self.draw_blank:
                self.set_untouchable()
                context.set_source_rgba(0.0, 0.0, 0.0, 0.0)
                context.set_operator(cairo.OPERATOR_SOURCE)
                context.paint()
                return
            else:
                (width, height) = self.get_size()
                surface = cairo.ImageSurface(
                    cairo.FORMAT_ARGB32, width, height)
                surface_ctx = cairo.Context(surface)
                self.overlay_draw(None, surface_ctx)
                reg = Gdk.cairo_region_create_from_surface(surface)
                self.input_shape_combine_region(reg)

        self.overlay_draw(_w, context, data)

    def overlay_draw(self, _w, context, data=None):
        """
        Draw overlay
        """

    def set_font(self, font):
        """
        Set the font used by the overlay
        """
        if self.text_font != font:
            self.text_font = font
            self.set_needs_redraw()

    def set_floating(self, floating, pos_x, pos_y, width, height):
        """
        Set if the window is floating and what dimensions to use
        """
        if width > 1.0 and height > 1.0:
            # Old data.
            (screen_x, screen_y, screen_width,
             screen_height) = self.get_display_coords()
            pos_x = float(pos_x) / screen_width
            pos_y = float(pos_y) / screen_height
            width = float(width) / screen_width
            height = float(height) / screen_height

        if self.floating != floating or self.pos_x != pos_x or self.pos_y != pos_y or self.width != width or self.height != height:
            # Special case for Cinnamon desktop : see https://github.com/trigg/Discover/issues/322
            if 'XDG_SESSION_DESKTOP' in os.environ and os.environ['XDG_SESSION_DESKTOP'] == 'cinnamon':
                floating = True

            self.floating = floating
            self.pos_x = pos_x
            self.pos_y = pos_y
            self.width = width
            self.height = height
            self.force_location()

    def set_untouchable(self):
        """
        Create a custom input shape and tell it that all of the window is a cut-out
        This allows us to have a window above everything but that never gets clicked on
        """
        (width, height) = self.get_size()
        surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, width, height)
        surface_ctx = cairo.Context(surface)
        surface_ctx.set_source_rgba(0.0, 0.0, 0.0, 0.0)
        surface_ctx.set_operator(cairo.OPERATOR_SOURCE)
        surface_ctx.paint()
        reg = Gdk.cairo_region_create_from_surface(surface)
        self.input_shape_combine_region(reg)

    def set_hide_on_mouseover(self, hide):
        if self.hide_on_mouseover != hide:
            self.hide_on_mouseover = hide
            if self.hide_on_mouseover:
                self.set_needs_redraw()
            else:
                self.set_untouchable()

    def set_mouseover_timer(self, time):
        self.timeout_mouse_over = time

    def unset_shape(self):
        """
        Remove XShape (not input shape)
        """
        if self.get_window():
            self.get_window().shape_combine_region(None, 0, 0)

    def force_location(self):
        """
        On X11 enforce the location and sane defaults
        On Wayland just store for later
        On Gamescope enforce size of display but only if it's the primary overlay
        """
        if self.discover.steamos and not self.piggyback_parent:
            (floating_x, floating_y, floating_width,
             floating_height) = self.get_floating_coords()
            self.resize(floating_width, floating_height)
            self.set_needs_redraw()
            return
        if not self.is_wayland:
            self.set_decorated(False)
            self.set_keep_above(True)

            (floating_x, floating_y, floating_width,
             floating_height) = self.get_floating_coords()
            self.resize(floating_width, floating_height)
            self.move(floating_x, floating_y)

        self.set_needs_redraw()

    def get_display_coords(self):
        if self.piggyback_parent:
            return self.piggyback_parent.get_display_coords()
        display = Gdk.Display.get_default()
        if "get_monitor" in dir(display):
            if self.monitor == None or self.monitor < 0:
                monitor = display.get_monitor(0)
            else:
                monitor = display.get_monitor(self.monitor)
            if monitor:
                geometry = monitor.get_geometry()
                return (geometry.x, geometry.y, geometry.width, geometry.height)
        log.warn("No monitor found! This is going to go badly")
        return (0, 0, 1920, 1080)  # We're in trouble

    def get_floating_coords(self):
        (screen_x, screen_y, screen_width, screen_height) = self.get_display_coords()
        if self.floating:
            if self.pos_x == None or self.pos_y == None or self.width == None or self.height == None:
                log.error("No usable floating position")

            if not self.is_wayland:
                return (screen_x + self.pos_x * screen_width, screen_y + self.pos_y * screen_height, self.width * screen_width, self.height * screen_height)
            return (self.pos_x * screen_width, self.pos_y * screen_height, self.width * screen_width, self.height * screen_height)
        else:
            return (0, 0, screen_width, screen_height)

    def set_needs_redraw(self, be_pushy=False):
        if (not self.hidden and self.enabled) or be_pushy:
            if self.piggyback_parent:
                self.piggyback_parent.set_needs_redraw(be_pushy=True)

            if self.redraw_id == None:
                self.redraw_id = GLib.idle_add(self.redraw)
            else:
                log.debug("Already awaiting paint")

            # If this overlay has data that expires after draw, plan for that here
            if self.timer_after_draw != None:
                GLib.timeout_add_seconds(self.timer_after_draw, self.redraw)

    def redraw(self):
        """
        Request a redraw.
        If we're using XShape (optionally or forcibly) then render the image into the shape
        so that we only cut out clear sections
        """
        self.redraw_id = None
        gdkwin = self.get_window()
        if self.piggyback_parent:
            self.piggyback_parent.redraw()
            return
        if gdkwin:
            compositing = self.get_screen().is_composited()
            if not compositing or self.force_xshape:
                (width, height) = self.get_size()
                surface = cairo.ImageSurface(
                    cairo.FORMAT_ARGB32, width, height)
                surface_ctx = cairo.Context(surface)
                self.overlay_draw(None, surface_ctx)
                reg = Gdk.cairo_region_create_from_surface(surface)
                gdkwin.shape_combine_region(reg, 0, 0)
            else:
                gdkwin.shape_combine_region(None, 0, 0)
        self.queue_draw()
        self.redraw_id = None
        return False

    def set_hidden(self, hidden):
        self.hidden = hidden
        self.set_enabled(self.enabled)

    def set_monitor(self, idx=None):
        """
        Set the monitor this overlay should display on.
        """
        if type(idx) is str:
            idx = 0
        if self.monitor != idx:
            self.monitor = idx
            if self.is_wayland:
                display = Gdk.Display.get_default()
                if "get_monitor" in dir(display):
                    monitor = display.get_monitor(self.monitor)
                    if monitor:
                        GtkLayerShell.set_monitor(self, monitor)
                    else:
                        self.hide()
                        self.set_wayland_state()
                        self.show()
                else:
                    log.error("No get_monitor in display")
                self.set_untouchable()
            self.force_location()
            self.set_needs_redraw()

    def set_align_x(self, align_right):
        """
        Set the alignment (True for right, False for left)
        """
        if self.align_right != align_right:
            self.align_right = align_right
            self.force_location()
            self.set_needs_redraw()

    def set_align_y(self, align_vert):
        """
        Set the veritcal alignment
        """
        if self.align_vert != align_vert:
            self.align_vert = align_vert
            self.force_location()
            self.set_needs_redraw()

    def col(self, col, alpha=1.0):
        """
        Convenience function to set the cairo context next colour
        """
        self.context.set_source_rgba(col[0], col[1], col[2], col[3] * alpha)

    def set_force_xshape(self, force):
        """
        Set if XShape should be forced
        """
        self.force_xshape = force

        if self.is_wayland or self.discover.steamos:
            # Wayland and XShape are a bad idea unless you're a fan on artifacts
            self.force_xshape = False

    def set_enabled(self, enabled):
        """
        Set if this overlay should be visible
        """
        self.enabled = enabled
        if self.piggyback_parent or self.piggyback:
            self.set_needs_redraw()

            if not self.piggyback_parent:
                self.set_gamescope_xatom(1 if enabled else 0)
            return
        if enabled and not self.hidden:
            self.show_all()
            self.set_untouchable()
        else:
            self.hide()

    def set_task(self, visible):
        self.set_skip_pager_hint(not visible)
        self.set_skip_taskbar_hint(not visible)

    def check_composite(self, _a=None, _b=None):
        # Called when an X11 session switched compositing on or off
        self.redraw()

    def screen_changed(self, screen=None):
        self.set_monitor(self.monitor)

    def mouseover(self, a=None, b=None):
        self.draw_blank = True
        self.set_needs_redraw()
        return True

    def mouseout(self, a=None, b=None):
        GLib.timeout_add_seconds(self.timeout_mouse_over, self.mouseout_timed)
        return True

    def mouseout_timed(self, a=None, b=None):
        self.draw_blank = False
        self.set_needs_redraw()
        return False



#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Functions & Classes to assist image loading."""
import urllib
import threading
import logging
import gi
import requests
import cairo
import PIL
import PIL.Image as Image
import os
import io
import copy
gi.require_version('GdkPixbuf', '2.0')
gi.require_version("Gtk", "3.0")
# pylint: disable=wrong-import-position
from gi.repository import Gio, GdkPixbuf, Gtk  # nopep8

log = logging.getLogger(__name__)


class SurfaceGetter():
    """Download and decode image using PIL and store as a cairo surface"""

    def __init__(self, func, url, identifier, size):
        self.func = func
        self.identifier = identifier
        self.url = url
        self.size = size

    def get_url(self):
        """Downloads and decodes"""
        try:
            resp = requests.get(
                self.url, stream=True, headers={
                    'Referer': 'https://streamkit.discord.com/overlay/voice',
                    'User-Agent': 'Mozilla/5.0'
                }
            )
            raw = resp.raw
            image = Image.open(raw)
            (surface, mask) = from_pil(image)

            self.func(self.identifier, surface, mask)
        except requests.HTTPError:
            log.error("Unable to open %s", self.url)
        except requests.TooManyRedirects:
            log.error("Unable to open %s - Too many redirects", self.url)
        except requests.Timeout:
            log.error("Unable to open %s - Timeout", self.url)
        except requests.ConnectionError:
            log.error("Unable to open %s - Connection error", self.url)
        except ValueError:
            log.error("Unable to read %s", self.url)
        except TypeError:
            log.error("Unable to read %s", self.url)
        except PIL.UnidentifiedImageError:
            log.error("Unknown image type:  %s", self.url)

    def get_file(self):
        errors = []
        # Grab icon from icon theme
        icon_theme = Gtk.IconTheme.get_default()
        icon = icon_theme.choose_icon(
            [self.url, None], -1, Gtk.IconLookupFlags.NO_SVG)

        if icon:
            try:
                image = Image.open(icon.get_filename())
                (surface, mask) = from_pil(image)
                if surface:
                    self.func(self.identifier, surface, mask)
                    return
            except ValueError:
                errors.append("Value Error - Unable to read %s" % (mixpath))
            except TypeError:
                errors.append("Type Error - Unable to read %s" % (mixpath))
            except PIL.UnidentifiedImageError:
                errors.append("Unknown image type: %s" % (mixpath))
            except FileNotFoundError:
                errors.append("File not found: %s" % (mixpath))
        # Not found in theme, try some common locations
        locations = [os.path.expanduser('~/.local/'), '/usr/', '/app']
        for prefix in locations:
            mixpath = os.path.join(os.path.join(
                prefix, 'share/icons/hicolor/256x256/apps/'), self.url + ".png")
            image = None
            try:
                image = Image.open(mixpath)
            except ValueError:
                errors.append("Value Error - Unable to read %s" % (mixpath))
            except TypeError:
                errors.append("Type Error - Unable to read %s" % (mixpath))
            except PIL.UnidentifiedImageError:
                errors.append("Unknown image type: %s" % (mixpath))
            except FileNotFoundError:
                errors.append("File not found: %s" % (mixpath))
            if image:
                (surface, mask) = from_pil(image)
                if surface:
                    self.func(self.identifier, surface, mask)
                    return
        for error in errors:
            log.error(error)


def from_pil(image, alpha=1.0, format='BGRa'):
    """
    :param im: Pillow Image
    :param alpha: 0..1 alpha to add to non-alpha images
    :param format: Pixel format for output surface
    """
    arr = bytearray()
    mask = bytearray()
    if 'A' not in image.getbands():
        image.putalpha(int(alpha * 255.0))
        arr = bytearray(image.tobytes('raw', format))
        mask = arr
    else:
        arr = bytearray(image.tobytes('raw', format))
        mask = copy.deepcopy((arr))
        idx = 0
        while idx < len(arr):
            if arr[idx] > 0:
                mask[idx] = 255
            else:
                mask[idx] = 0
            # Cairo expects the raw data to be pre-multiplied alpha
            # This means when we change the alpha level we need to change the RGB channels equally
            arr[idx] = int(arr[idx] * alpha)
            idx += 1
    surface = cairo.ImageSurface.create_for_data(
        arr, cairo.FORMAT_ARGB32, image.width, image.height)
    mask = cairo.ImageSurface.create_for_data(
        mask, cairo.FORMAT_ARGB32, image.width, image.height)
    return (surface, mask)


def to_pil(surface):
    if surface.get_format() == cairo.Format.ARGB32:
        return Image.frombuffer('RGBA', (surface.get_width(), surface.get_height()), surface.get_data(), 'raw', "BGRA", surface.get_stride())
    return Image.frombuffer("RGB", (surface.get_width(), surface.get_height()), surface.get_data(), 'raw', "BGRX", stride)


def get_surface(func, identifier, ava, size):
    """Download to cairo surface"""
    image_getter = SurfaceGetter(func, identifier, ava, size)
    if identifier.startswith('http'):
        thread = threading.Thread(target=image_getter.get_url)
        thread.start()
    else:
        thread = threading.Thread(target=image_getter.get_file)
        thread.start()


def get_aspected_size(img, width, height, anchor=0, hanchor=0):
    """Get dimensions of image keeping current aspect ratio"""
    pic_width = img.get_width()
    pic_height = img.get_height()
    if pic_height < 1 or height < 1:
        return (0, 0, 0, 0)
    img_aspect = pic_width / pic_height
    rect_aspect = width / height

    offset_y = 0
    offset_x = 0
    if img_aspect > rect_aspect:
        old_height = height
        height = width / img_aspect
        if anchor == 0:
            offset_y = offset_y + (old_height - height)
        if anchor == 1:
            offset_y = offset_y + ((old_height - height) / 2)
    elif img_aspect < rect_aspect:
        old_width = width
        width = height * img_aspect
        if hanchor == 2:
            offset_x = offset_x + (old_width - width)
        if hanchor == 1:
            offset_x = offset_x + ((old_width - width) / 2)
    return (offset_x, offset_y, width, height)


def draw_img_to_rect(img, ctx,
                     pos_x, pos_y,
                     width, height,
                     path=False, aspect=False,
                     anchor=0, hanchor=0, alpha=1.0):
    """Draw cairo surface onto context

    Path - only add the path do not fill : True/False
    Aspect - keep aspect ratio : True/False
    Anchor - with aspect : 0=left 1=middle 2=right
    HAnchor - with apect : 0=bottom 1=middle 2=top
    """

    ctx.save()
    offset_x = 0
    offset_y = 0
    if aspect:
        (offset_x, offset_y, width, height) = get_aspected_size(
            img, width, height, anchor=anchor, hanchor=hanchor)

    ctx.translate(pos_x + offset_x, pos_y + offset_y)
    ctx.scale(width, height)
    ctx.scale(1 / img.get_width(), 1 / img.get_height())

    if alpha != 1.0:
        # Honestly, couldn't find a 'use-image-with-modifier' option
        # Tried RasterSourcePattern but it appears... broken? in the python implementation
        # Or just lacking documentation.

        # Pass raw data to PIL and then back with an alpha modifier
        ctx.set_source_surface(
            from_pil(
                to_pil(img),
                alpha
            )[0],
            0, 0)
    else:
        ctx.set_source_surface(img, 0, 0)

    ctx.rectangle(0, 0, img.get_width(), img.get_height())
    if not path:
        ctx.fill()
    ctx.restore()
    return (width, height)


def draw_img_to_mask(img, ctx,
                     pos_x, pos_y,
                     width, height,
                     path=False, aspect=False,
                     anchor=0, hanchor=0):
    """Draw cairo surface as mask into context

    Path - only add the path do not fill : True/False
    Aspect - keep aspect ratio : True/False
    Anchor - with aspect : 0=left 1=middle 2=right
    HAnchor - with apect : 0=bottom 1=middle 2=top
    """

    ctx.save()
    offset_x = 0
    offset_y = 0
    if aspect:
        (offset_x, offset_y, width, height) = get_aspected_size(
            img, width, height, anchor=anchor, hanchor=hanchor)

    ctx.translate(pos_x + offset_x, pos_y + offset_y)
    ctx.scale(width, height)
    ctx.scale(1 / img.get_width(), 1 / img.get_height())

    ctx.rectangle(0, 0, img.get_width(), img.get_height())
    if not path:
        ctx.mask_surface(img, 0, 0)
    ctx.restore()
    return (width, height)



#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Main application class"""
import gettext
import os
import time
import sys
import re
import traceback
import logging
import pkg_resources
import json
import signal
import gi
from configparser import ConfigParser

from .settings_window import MainSettingsWindow
from .voice_overlay import VoiceOverlayWindow
from .text_overlay import TextOverlayWindow
from .notification_overlay import NotificationOverlayWindow
from .discord_connector import DiscordConnector
from .audio_assist import DiscoverAudioAssist

gi.require_version("Gtk", "3.0")
# pylint: disable=wrong-import-position,wrong-import-order
from gi.repository import Gtk, GLib, Gio, Gdk  # nopep8

try:
    from xdg.BaseDirectory import xdg_config_home
except ModuleNotFoundError:
    from xdg import XDG_CONFIG_HOME as xdg_config_home

log = logging.getLogger(__name__)
t = gettext.translation(
    'default', pkg_resources.resource_filename('discover_overlay', 'locales'), fallback=True)
_ = t.gettext


class Discover:
    """Main application class"""

    def __init__(self, rpc_file, config_file, channel_file, debug_file, args):
        self.mix_settings = False
        self.ind = None
        self.tray = None
        self.steamos = False
        self.connection = None
        self.show_settings_delay = False
        self.settings = None

        self.debug_file = debug_file
        self.channel_file = channel_file
        self.config_file = config_file
        self.rpc_file = rpc_file

        self.do_args(args, True)
        if "GAMESCOPE_WAYLAND_DISPLAY" in os.environ:
            log.info(
                "GameScope session detected. Enabling steam and gamescope integration")
            self.steamos = True
            self.show_settings_delay = True
            self.mix_settings = True
            settings = Gtk.Settings.get_default()
            if settings:
                settings.set_property(
                    "gtk-application-prefer-dark-theme", Gtk.true)

        self.create_gui()

        self.connection = DiscordConnector(self)

        self.connection.connect()
        self.audio_assist = DiscoverAudioAssist(self)

        rpc_file = Gio.File.new_for_path(rpc_file)
        monitor = rpc_file.monitor_file(0, None)
        monitor.connect("changed", self.rpc_changed)

        config_file = Gio.File.new_for_path(config_file)
        monitor_config = config_file.monitor_file(0, None)
        monitor_config.connect("changed", self.config_changed)

        self.config_changed()

        Gtk.main()

    def do_args(self, data, normal_close):
        """
        Read in arg list from command or RPC and act accordingly
        """
        if "--help" in data or "-h" in data:
            print("%s: discover-overlay [OPTIONS]... " % (_("Usage")))
            print(_("Show an X11 or wlroots overlay with information"))
            print(_("from Discord client"))
            print("")
            print("  -c, --configure        ", _("Open configuration window"))
            print("  -x, --close            ",
                  _("Close currently running instance"))
            print("  -v, --debug            ",
                  _("Verbose output for aid in debugging"))
            print("  -h, --help             ", _("This screen"))
            print("      --hide             ", _("Hide overlay"))
            print("      --show             ", _("Show overlay"))
            print("      --rpc              ",
                  _("Send command, not start new instance."))
            print("      --mute             ", _("Set own user to mute"))
            print("      --unmute           ", _("Set unmuted"))
            print("      --deaf             ", _("Set own user to deafened"))
            print("      --undeaf           ", _("Unset user deafened state"))
            print("      --moveto=XX        ",
                  _("Move the user into voice room, by Room ID"))
            print("      --minimized        ",
                  _("If tray icon is enabled, start with only tray icon and no configuration window"))
            print("")
            print(_("For gamescope compatibility ensure ENV has 'GDK_BACKEND=x11'"))
            if normal_close:
                sys.exit(0)
        if "--close" in data or "-x" in data:
            sys.exit(0)
        if "--steamos" in data or "-s" in data:
            self.steamos = True
        if "--hide" in data:
            self.config_set("general", "hideoverlay", "True")
        if "--show" in data:
            self.config_set("general", "hideoverlay", "False")
        if "--mute" in data:
            if self.connection:
                self.connection.set_mute(True)
        if "--unmute" in data:
            if self.connection:
                self.connection.set_mute(False)
        if "--deaf" in data:
            if self.connection:
                self.connection.set_deaf(True)
        if "--undeaf" in data:
            if self.connection:
                self.connection.set_deaf(False)
        if "--refresh-guilds" in data:
            if self.connection:
                self.connection.req_guilds()
        pattern = re.compile("--moveto=([0-9]+)")
        if any((match := pattern.match(x)) for x in data):
            if self.connection:
                self.connection.change_voice_room(match.group(1))
        guild_pattern = re.compile("--guild-request=([0-9]+)")
        if any((match := guild_pattern.match(x)) for x in data):
            if self.connection:
                self.connection.request_text_rooms_for_guild(match.group(1))

    def config_set(self, context, key, value):
        config = self.config()
        if not context in config.sections():
            config.add_section(context)
        config.set(context, key, value)
        with open(self.config_file, 'w') as file:
            config.write(file)

    def config(self):
        config = ConfigParser(interpolation=None)
        config.read(self.config_file)
        return config

    def rpc_changed(self, _a=None, _b=None, _c=None, _d=None):
        """
        Called when the RPC file has been altered
        """
        with open(self.rpc_file, "r") as tfile:
            data = tfile.readlines()
            if len(data) >= 1:
                self.do_args(data[0].strip().split(" "), False)

    def config_changed(self, _a=None, _b=None, _c=None, _d=None):
        """
        Called when the config file has been altered
        """
        # Read new config
        config = self.config()

        # Set Voice overlay options
        self.voice_overlay.set_align_x(config.getboolean(
            "main", "rightalign", fallback=False))
        self.voice_overlay.set_align_y(
            config.getint("main", "topalign", fallback=1))
        self.voice_overlay.set_bg(json.loads(config.get(
            "main", "bg_col", fallback="[0.0,0.0,0.0,0.5]")))
        self.voice_overlay.set_fg(json.loads(config.get(
            "main", "fg_col", fallback="[1.0,1.0,1.0,1.0]")))
        self.voice_overlay.set_fg_hi(json.loads(config.get(
            "main", "fg_hi_col", fallback="[1.0,1.0,1.0,1.0]")))
        self.voice_overlay.set_tk(json.loads(config.get(
            "main", "tk_col", fallback="[0.0,0.7,0.0,1.0]")))
        self.voice_overlay.set_mt(json.loads(config.get(
            "main", "mt_col", fallback="[0.6,0.0,0.0,1.0]")))
        self.voice_overlay.set_mute_bg(json.loads(config.get(
            "main", "mt_bg_col", fallback="[0.0,0.0,0.0,0.5]")))
        self.voice_overlay.set_hi(json.loads(config.get(
            "main", "hi_col", fallback="[0.0,0.0,0.0,0.5]")))
        self.voice_overlay.set_bo(json.loads(config.get(
            "main", "bo_col", fallback="[0.0,0.0,0.0,0.0]")))
        self.voice_overlay.set_avatar_bg_col(json.loads(config.get(
            "main", "avatar_bg_col", fallback="[0.0,0.0,0.0,0.0]")))
        self.voice_overlay.set_avatar_size(
            config.getint("main", "avatar_size", fallback=48))
        self.voice_overlay.set_nick_length(
            config.getint("main", "nick_length", fallback=32))
        self.voice_overlay.set_icon_spacing(
            config.getint("main", "icon_spacing", fallback=8))
        self.voice_overlay.set_text_padding(
            config.getint("main", "text_padding", fallback=6))
        self.voice_overlay.set_text_baseline_adj(config.getint(
            "main", "text_baseline_adj", fallback=0))
        font = config.get("main", "font", fallback=None)
        title_font = config.get("main", "title_font", fallback=None)
        self.voice_overlay.set_square_avatar(config.getboolean(
            "main", "square_avatar", fallback=True))
        self.voice_overlay.set_only_speaking(config.getboolean(
            "main", "only_speaking", fallback=False))
        self.voice_overlay.set_only_speaking_grace_period(config.getint(
            "main", "only_speaking_grace", fallback=0))
        self.voice_overlay.set_highlight_self(config.getboolean(
            "main", "highlight_self", fallback=False))
        self.voice_overlay.set_icon_only(config.getboolean(
            "main", "icon_only", fallback=False))
        monitor = 0
        try:
            monitor = config.getint("main", "monitor", fallback=0)
        except:
            pass
        self.voice_overlay.set_vert_edge_padding(config.getint(
            "main", "vert_edge_padding", fallback=0))
        self.voice_overlay.set_horz_edge_padding(config.getint(
            "main", "horz_edge_padding", fallback=0))
        floating = config.getboolean("main", "floating", fallback=False)
        floating_x = config.getfloat("main", "floating_x", fallback=0.0)
        floating_y = config.getfloat("main", "floating_y", fallback=0.0)
        floating_w = config.getfloat("main", "floating_w", fallback=0.1)
        floating_h = config.getfloat("main", "floating_h", fallback=0.1)
        self.voice_overlay.set_order(
            config.getint("main", "order", fallback=0))
        self.voice_overlay.set_hide_on_mouseover(
            config.getboolean("main", "autohide", fallback=False))
        self.voice_overlay.set_mouseover_timer(
            config.getint("main", "autohide_timer", fallback=1))

        self.voice_overlay.set_horizontal(config.getboolean(
            "main", "horizontal", fallback=False))
        self.voice_overlay.set_guild_ids(self.parse_guild_ids(
            config.get("main", "guild_ids", fallback="")))
        self.voice_overlay.set_overflow(
            config.getint("main", "overflow", fallback=0))
        self.voice_overlay.set_show_connection(config.getboolean(
            "main", "show_connection", fallback=False))
        self.voice_overlay.set_show_title(config.getboolean(
            "main", "show_title", fallback=False))
        self.voice_overlay.set_show_disconnected(config.getboolean(
            "main", "show_disconnected", fallback=False))
        self.voice_overlay.set_border_width(
            config.getint("main", "border_width", fallback=2))
        self.voice_overlay.set_icon_transparency(config.getfloat(
            "main", "icon_transparency", fallback=1.0))
        self.voice_overlay.set_show_avatar(
            config.getboolean("main", "show_avatar", fallback=True))
        self.voice_overlay.set_fancy_border(config.getboolean("main",
                                                              "fancy_border", fallback=True))
        self.voice_overlay.set_show_dummy(config.getboolean("main",
                                                            "show_dummy", fallback=False))
        self.voice_overlay.set_dummy_count(config.getint("main",
                                                         "dummy_count", fallback=10))

        self.voice_overlay.set_monitor(monitor)

        self.voice_overlay.set_enabled(True)

        self.voice_overlay.set_floating(
            floating, floating_x, floating_y, floating_w, floating_h)

        if font:
            self.voice_overlay.set_font(font)
        if title_font:
            self.voice_overlay.set_title_font(title_font)

        self.voice_overlay.set_fade_out_inactive(
            config.getboolean("main", "fade_out_inactive", fallback=False),
            config.getint("main", "inactive_time", fallback=10),
            config.getint("main", "inactive_fade_time", fallback=30),
            config.getfloat("main", "fade_out_limit", fallback=0.3)
        )

        # Set Text overlay options
        self.text_overlay.set_enabled(config.getboolean(
            "text", "enabled", fallback=False))
        self.text_overlay.set_align_x(config.getboolean(
            "text", "rightalign", fallback=True))
        self.text_overlay.set_align_y(
            config.getint("text", "topalign", fallback=2))
        monitor = 0
        try:
            monitor = config.getint("text", "monitor", fallback=0)
        except:
            pass
        floating = config.getboolean("text", "floating", fallback=True)
        floating_x = config.getfloat("text", "floating_x", fallback=0.0)
        floating_y = config.getfloat("text", "floating_y", fallback=0.0)
        floating_w = config.getfloat("text", "floating_w", fallback=0.1)
        floating_h = config.getfloat("text", "floating_h", fallback=0.1)

        channel = config.get("text", "channel", fallback="0")
        guild = config.get("text", "guild", fallback="0")
        self.connection.set_text_channel(channel, guild)

        self.font = config.get("text", "font", fallback=None)
        self.text_overlay.set_bg(json.loads(config.get(
            "text", "bg_col", fallback="[0.0,0.0,0.0,0.5]")))
        self.text_overlay.set_fg(json.loads(config.get(
            "text", "fg_col", fallback="[1.0,1.0,1.0,1.0]")))
        self.text_overlay.set_popup_style(config.getboolean(
            "text", "popup_style", fallback=False))
        self.text_overlay.set_text_time(
            config.getint("text", "text_time", fallback=30))
        self.text_overlay.set_show_attach(config.getboolean(
            "text", "show_attach", fallback=True))
        self.text_overlay.set_line_limit(
            config.getint("text", "line_limit", fallback=20))
        self.text_overlay.set_hide_on_mouseover(
            config.getboolean("text", "autohide", fallback=False))
        self.text_overlay.set_mouseover_timer(
            config.getint("text", "autohide_timer", fallback=1))

        self.text_overlay.set_monitor(monitor)
        self.text_overlay.set_floating(
            floating, floating_x, floating_y, floating_w, floating_h)

        if self.font:
            self.text_overlay.set_font(self.font)

        # Set Notification overlay options
        self.notification_overlay.set_enabled(config.getboolean(
            "notification", "enabled", fallback=False))
        self.notification_overlay.set_align_x(config.getboolean(
            "notification", "rightalign", fallback=True))
        self.notification_overlay.set_align_y(
            config.getint("notification", "topalign", fallback=2))
        monitor = 0
        try:
            monitor = config.getint("notification", "monitor", fallback=0)
        except:
            pass
        floating = config.getboolean(
            "notification", "floating", fallback=False)
        floating_x = config.getfloat(
            "notification", "floating_x", fallback=0.0)
        floating_y = config.getfloat(
            "notification", "floating_y", fallback=0.0)
        floating_w = config.getfloat(
            "notification", "floating_w", fallback=0.1)
        floating_h = config.getfloat(
            "notification", "floating_h", fallback=0.1)
        font = config.get("notification", "font", fallback=None)
        self.notification_overlay.set_bg(json.loads(config.get(
            "notification", "bg_col", fallback="[0.0,0.0,0.0,0.5]")))
        self.notification_overlay.set_fg(json.loads(config.get(
            "notification", "fg_col", fallback="[1.0,1.0,1.0,1.0]")))
        self.notification_overlay.set_text_time(config.getint(
            "notification", "text_time", fallback=10))
        self.notification_overlay.set_show_icon(config.getboolean(
            "notification", "show_icon", fallback=True))
        self.notification_overlay.set_reverse_order(config.getboolean(
            "notification", "rev", fallback=False))
        self.notification_overlay.set_limit_width(config.getint(
            "notification", "limit_width", fallback=400))
        self.notification_overlay.set_icon_left(config.getboolean(
            "notification", "icon_left", fallback=True))
        self.notification_overlay.set_icon_pad(config.getint(
            "notification", "icon_padding", fallback=8))
        self.notification_overlay.set_icon_size(config.getint(
            "notification", "icon_size", fallback=32))
        self.notification_overlay.set_padding(config.getint(
            "notification", "padding", fallback=8))
        self.notification_overlay.set_border_radius(config.getint(
            "notification", "border_radius", fallback=8))
        self.notification_overlay.set_testing(config.getboolean(
            "notification", "show_dummy", fallback=False))
        self.font = config.get("notification", "font", fallback=None)

        if self.font:
            self.notification_overlay.set_font(self.font)

        self.notification_overlay.set_monitor(monitor)
        self.notification_overlay.set_floating(
            floating, floating_x, floating_y, floating_w, floating_h)
        if self.font:
            self.notification_overlay.set_font(self.font)

        # Set Core settings
        self.set_force_xshape(
            config.getboolean("general", "xshape", fallback=False))

        hidden = config.getboolean("general", "hideoverlay", fallback=False)
        self.voice_overlay.set_hidden(hidden)
        self.text_overlay.set_hidden(hidden)
        self.notification_overlay.set_hidden(hidden)

        self.audio_assist.set_enabled(config.getboolean(
            "general", "audio_assist", fallback=False))

    def parse_guild_ids(self, guild_ids_str):
        """Parse the guild_ids from a str and return them in a list"""
        guild_ids = []
        for guild_id in guild_ids_str.split(","):
            guild_id = guild_id.strip()
            if guild_id != "":
                guild_ids.append(guild_id)
        return guild_ids

    def create_gui(self):
        """
        Create Systray & associated menu, overlays & settings windows
        """
        self.voice_overlay = VoiceOverlayWindow(self)

        if self.steamos:
            self.text_overlay = TextOverlayWindow(self, self.voice_overlay)
            self.notification_overlay = NotificationOverlayWindow(
                self, self.text_overlay)
        else:
            self.text_overlay = TextOverlayWindow(self)
            self.notification_overlay = NotificationOverlayWindow(self)

        if self.mix_settings:
            MainSettingsWindow(
                self.config_file, self.rpc_file, self.channel_file, [])

    def toggle_show(self, _obj=None):
        if self.voice_overlay:
            hide = not self.voice_overlay.hidden
            self.voice_overlay.set_hidden(hide)
            if self.text_overlay:
                self.text_overlay.set_hidden(hide)
            if self.notification_overlay:
                self.notification_overlay.set_hidden(hide)

    def close(self, _a=None, _b=None, _c=None):
        """
        End of the program
        """
        Gtk.main_quit()

    def set_force_xshape(self, force):
        """
        Set if XShape should be forced
        """
        self.voice_overlay.set_force_xshape(force)
        if self.text_overlay:
            self.text_overlay.set_force_xshape(force)
        if self.notification_overlay:
            self.notification_overlay.set_force_xshape(force)

    def set_show_task(self, visible):
        if self.voice_overlay:
            self.voice_overlay.set_task(visible)
        if self.text_overlay:
            self.text_overlay.set_task(visible)
        if self.notification_overlay:
            self.notification_overlay.set_task(visible)

    def set_mute_async(self, mute):
        if mute != None:
            GLib.idle_add(self.connection.set_mute, mute)

    def set_deaf_async(self, deaf):
        if deaf != None:
            GLib.idle_add(self.connection.set_deaf, deaf)


def entrypoint():
    """
    Entry Point.

    Find all needed file locations and read args

    if '--rpc' simply pass them over the rpc file

    if '-c' or '--configure' start the config window only

    otherwise start overlay
    """

    signal.signal(signal.SIGINT, signal.SIG_DFL)
    # Find Config directory
    config_dir = os.path.join(xdg_config_home, "discover_overlay")
    os.makedirs(config_dir, exist_ok=True)

    # Find RPC, Channel info, config and debug files
    rpc_file = os.path.join(config_dir, "discover_overlay.rpc")
    channel_file = os.path.join(config_dir, "channels.rpc")
    config_file = os.path.join(config_dir, "config.ini")
    debug_file = os.path.join(config_dir, "output.txt")

    # Prepare logger
    logging.getLogger().setLevel(logging.INFO)
    FORMAT = "%(levelname)s - %(name)s - %(message)s"
    if "--debug" in sys.argv or "-v" in sys.argv:
        logging.getLogger().setLevel(logging.DEBUG)
        logging.basicConfig(filename=debug_file, format=FORMAT)
    else:
        logging.basicConfig(format=FORMAT)
    log = logging.getLogger(__name__)
    log.info("Starting Discover Overlay: %s",
             pkg_resources.get_distribution('discover_overlay').version)

    # Hedge against the bet gamescope ships with some WAYLAND_DISPLAY
    # Compatibility and we're not ready yet
    if 'GAMESCOPE_WAYLAND_DISPLAY' in os.environ:
        os.unsetenv("WAYLAND_DISPLAY")

    # Catch any errors and log them
    try:
        if "--rpc" in sys.argv:
            # Send command to overlay
            line = ""
            for arg in sys.argv[1:]:
                line = "%s %s" % (line, arg)
            with open(rpc_file, "w") as tfile:
                tfile.write(line)
                log.warning("Sent RPC command")
        else:
            if "-c" in sys.argv or "--configure" in sys.argv:
                # Show config window
                settings = MainSettingsWindow(
                    config_file, rpc_file, channel_file, sys.argv[1:])
                Gtk.main()
            else:
                # Tell any other running overlay to close
                with open(rpc_file, "w") as tfile:
                    tfile.write("--close")
                # Show the overlay
                Discover(rpc_file, config_file, channel_file,
                         debug_file, sys.argv[1:])
        return

    except Exception as ex:
        log.error(ex)
        log.error(traceback.format_exc())
        sys.exit(1)



#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
The connector for discord.
Connects in as if it was Streamkit for OBS or Xsplit and
communicates to get voice & text info to display

Terminology:
GUILDS - Often called 'Servers' in discord. It is the group of users and channels that make up
         one server.
CHANNEL - Often called 'Rooms'. Both voice and text channels are types of channel
"""
import select
import time
import json
import sys
import logging
import calendar
import websocket
import requests

import gi
from gi.repository import GLib

log = logging.getLogger(__name__)


class DiscordConnector:
    """
    The connector for discord.
    Connects in as if it was Streamkit for OBS or Xsplit and
    communicates to get voice & text info to display
    """

    def __init__(self, discover):
        self.discover = discover
        self.websocket = None
        self.access_token = discover.config().get(
            "cache", "access_token", fallback=None)
        self.oauth_token = "207646673902501888"

        self.guilds = {}
        self.channels = {}
        self.user = {}
        self.userlist = {}
        self.in_room = []
        self.current_guild = "0"
        self.current_voice = "0"
        self.current_text = "0"
        self.current_text_guild = "0"
        self.list_altered = False
        self.text_altered = False
        self.text = []
        self.authed = False
        self.last_rate_limit_send = 0

        self.socket_watch = None

        self.rate_limited_channels = []
        self.reconnect_cb = None

    def get_access_token_stage1(self):
        """
        First stage of getting an access token. Request authorization from Discord client
        """
        if self.access_token:
            self.req_auth()
            return
        cmd = {
            "cmd": "AUTHORIZE",
            "args":
            {
                "client_id": self.oauth_token,
                "scopes": ["rpc", "messages.read", "rpc.notifications.read"],
                "prompt": "none",
            },
            "nonce": "deadbeef"
        }
        self.websocket.send(json.dumps(cmd))

    def get_access_token_stage2(self, code1):
        """
        Second stage of getting an access token. Give auth code to streamkit
        """
        url = "https://streamkit.discord.com/overlay/token"
        myobj = {"code": code1}
        response = requests.post(url, json=myobj)
        try:
            jsonresponse = json.loads(response.text)
        except json.JSONDecodeError:
            jsonresponse = {}
        if "access_token" in jsonresponse:
            self.access_token = jsonresponse["access_token"]
            self.req_auth()
        else:
            log.error("No access token in json response")
            log.error(response.text)
            log.error("The user most likely denied permission for this app")
            sys.exit(1)

    def set_channel(self, channel, guild, need_req=True):
        """
        Set currently active voice channel
        """
        if not channel:
            if self.current_voice:
                self.unsub_voice_channel(self.current_voice)
            self.current_voice = "0"
            self.current_guild = "0"
            self.discover.voice_overlay.set_blank()
            self.in_room = []
            return
        if channel != self.current_voice:
            if self.current_voice != "0":
                self.unsub_voice_channel(self.current_voice)
            self.sub_voice_channel(channel)
            self.current_voice = channel
            self.current_guild = guild
            if need_req:
                self.req_channel_details(channel)

    def set_text_channel(self, channel, guild, need_req=True):
        """
        Set currently active text channel
        """
        if not channel:
            self.current_text = "0"
            self.current_text_guild = "0"
            return
        if guild != self.current_text_guild:
            self.current_text_guild = guild
            self.request_text_rooms_for_guild(guild)
        if channel != self.current_text:
            self.current_text = channel
            self.current_text_guild = guild
            self.start_listening_text(channel)
            if need_req:
                self.req_channel_details(channel)

    def set_in_room(self, userid, present):
        """
        Set user currently in given room
        """
        if present:
            if userid not in self.in_room:
                self.in_room.append(userid)
        else:
            if userid in self.in_room:
                self.in_room.remove(userid)

    def add_text(self, message):
        """
        Add line of text to text list. Assumes the message is from the correct room
        """
        utc_time = None
        try:
            utc_time = time.strptime(
                message["timestamp"], "%Y-%m-%dT%H:%M:%S.%f%z")
        except ValueError:
            utc_time = time.strptime(
                message["timestamp"], "%Y-%m-%dT%H:%M:%S%z")

        epoch_time = calendar.timegm(utc_time)
        username = message["author"]["username"]
        if ("nick" in message and message['nick'] and len(message["nick"]) > 1
                and 'object Object' not in json.dumps(message["nick"])):
            username = message["nick"]
        colour = "#ffffff"
        if "author_color" in message:
            colour = message["author_color"]

        self.text.append({'id': message["id"],
                          'content': self.get_message_from_message(message),
                          'nick': username,
                          'nick_col': colour,
                          'time': epoch_time,
                          'attach': self.get_attachment_from_message(message),
                          })
        self.text_altered = True

    def update_text(self, message_in):
        """
        Update a line of text
        """
        for idx in range(0, len(self.text)):
            message = self.text[idx]
            if message['id'] == message_in['id']:
                new_message = {'id': message['id'],
                               'content': self.get_message_from_message(message_in),
                               'nick': message['nick'],
                               'nick_col': message['nick_col'],
                               'time': message['time'],
                               'attach': message['attach']}
                self.text[idx] = new_message
                self.text_altered = True
                return

    def delete_text(self, message_in):
        """
        Delete a line of text
        """
        for idx in range(0, len(self.text)):
            message = self.text[idx]
            if message['id'] == message_in['id']:
                del self.text[idx]
                self.text_altered = True
                return

    def get_message_from_message(self, message):
        """
        Messages are sent as JSON objects, with varying information.
        Decides which bits are shown and which are discarded
        """
        if "content_parsed" in message:
            return message["content_parsed"]
        elif "content" in message and len(message["content"]) > 0:
            return message["content"]
        elif "embeds" in message and len(message["embeds"]) == 1:
            if "rawDescription" in message["embeds"][0]:
                return message["embeds"][0]["rawDescription"]
            if "author" in message["embeds"][0]:
                return message["embeds"][0]["author"]["name"]
        elif "attachments" in message and len(message["attachments"]) == 1:
            return ""
        return ""

    def get_attachment_from_message(self, message):
        """
        Messages with attachments come in different forms, decide what is and is
        not an attachment
        """
        if "attachments" in message and len(message["attachments"]) == 1:
            return message["attachments"]
        return None

    def update_user(self, user):
        """
        Update user information
        Pass along our custom user information from version to version
        """
        if user["id"] in self.userlist:
            olduser = self.userlist[user["id"]]
            if "mute" not in user and "mute" in olduser:
                user["mute"] = olduser["mute"]
            if "deaf" not in user and "deaf" in olduser:
                user["deaf"] = olduser["deaf"]
            if "speaking" not in user and "speaking" in olduser:
                user["speaking"] = olduser["speaking"]
            if "nick" not in user and "nick" in olduser:
                user["nick"] = olduser["nick"]
            if "lastspoken" not in user and "lastspoken" in olduser:
                user["lastspoken"] = olduser["lastspoken"]
            if olduser["avatar"] != user["avatar"]:
                self.discover.voice_overlay.delete_avatar(user["id"])
        if "lastspoken" not in user:  # Still nothing?
            user["lastspoken"] = 0  # EEEEPOOCH EEEEEPOCH! BELIEVE MEEEE
        if "speaking" not in user:
            user["speaking"] = False
        self.userlist[user["id"]] = user

    def on_message(self, message):
        """
        Recieve websocket message super-function
        """
        j = json.loads(message)
        if j["cmd"] == "AUTHORIZE":
            if 'data' in j and 'code' in j['data']:
                self.get_access_token_stage2(j["data"]["code"])
            else:
                log.error("Authorization rejected")
                sys.exit(0)
            return
        elif j["cmd"] == "DISPATCH":
            if j["evt"] == "READY":
                self.req_auth()
            elif j["evt"] == "VOICE_STATE_UPDATE":
                self.list_altered = True
                thisuser = j["data"]["user"]
                nick = j["data"]["nick"]
                thisuser["nick"] = nick
                mute = (j["data"]["voice_state"]["mute"] or
                        j["data"]["voice_state"]["self_mute"] or
                        j["data"]["voice_state"]["suppress"])
                deaf = j["data"]["voice_state"]["deaf"] or j["data"]["voice_state"]["self_deaf"]
                thisuser["mute"] = mute
                thisuser["deaf"] = deaf
                if self.current_voice != "0":
                    self.update_user(thisuser)
                self.set_in_room(thisuser["id"], True)
            elif j["evt"] == "VOICE_STATE_CREATE":
                self.list_altered = True
                thisuser = j["data"]["user"]
                nick = j["data"]["nick"]
                thisuser["nick"] = nick
                self.update_user(thisuser)
                # We've joined a room... but where?
                if j["data"]["user"]["id"] == self.user["id"]:
                    self.find_user()
            elif j["evt"] == "VOICE_STATE_DELETE":
                self.list_altered = True
                self.set_in_room(j["data"]["user"]["id"], False)
                if j["data"]["user"]["id"] == self.user["id"]:
                    self.in_room = []
                    self.find_user()
                    self.discover.voice_overlay.set_channel_title(None)
                    self.discover.voice_overlay.set_channel_icon(None)
                    # User might have been forcibly moved room
            elif j["evt"] == "SPEAKING_START":
                self.list_altered = True
                self.userlist[j["data"]["user_id"]]["speaking"] = True
                self.userlist[j["data"]["user_id"]]["lastspoken"] = time.time()
                self.set_in_room(j["data"]["user_id"], True)
            elif j["evt"] == "SPEAKING_STOP":
                self.list_altered = True
                if j["data"]["user_id"] in self.userlist:
                    self.userlist[j["data"]["user_id"]]["speaking"] = False
                self.set_in_room(j["data"]["user_id"], True)
            elif j["evt"] == "VOICE_CHANNEL_SELECT":
                if j["data"]["channel_id"]:
                    self.set_channel(j["data"]["channel_id"],
                                     j["data"]["guild_id"])
                else:
                    self.set_channel(None, None)
            elif j["evt"] == "VOICE_CONNECTION_STATUS":
                self.discover.voice_overlay.set_connection_status(j["data"])
            elif j["evt"] == "MESSAGE_CREATE":
                if self.current_text == j["data"]["channel_id"]:
                    self.add_text(j["data"]["message"])
            elif j["evt"] == "MESSAGE_UPDATE":
                if self.current_text == j["data"]["channel_id"]:
                    self.update_text(j["data"]["message"])
            elif j["evt"] == "MESSAGE_DELETE":
                if self.current_text == j["data"]["channel_id"]:
                    self.delete_text(j["data"]["message"])
            elif j["evt"] == "CHANNEL_CREATE":
                # We haven't been told what guild this is in
                self.req_channel_details(j["data"]["id"], 'new')
            elif j["evt"] == "NOTIFICATION_CREATE":
                self.discover.notification_overlay.add_notification_message(j)
            elif j["evt"] == "VOICE_SETTINGS_UPDATE":
                source = j['data']['input']['device_id']
                sink = j['data']['output']['device_id']
                if sink == 'default':
                    for available_sink in j['data']['output']['available_devices']:
                        if available_sink['id'] == 'default':
                            sink = available_sink['name'][9:]
                if source == 'default':
                    for available_source in j['data']['input']['available_devices']:
                        if available_source['id'] == 'default':
                            source = available_source['name'][9:]
                self.discover.audio_assist.set_devices(sink, source)

            else:
                log.warning(j)
            return
        elif j["cmd"] == "AUTHENTICATE":
            if j["evt"] == "ERROR":
                self.access_token = None
                self.get_access_token_stage1()
                return
            else:
                self.discover.config_set(
                    "cache", "access_token", self.access_token)
                self.req_guilds()
                self.user = j["data"]["user"]
                log.info(
                    "ID is %s", self.user["id"])
                log.info(
                    "Logged in as %s", self.user["username"])
                self.authed = True
                self.on_connected()
                return
        elif j["cmd"] == "GET_GUILDS":
            for guild in j["data"]["guilds"]:
                self.guilds[guild["id"]] = guild
                self.dump_channel_data()
            return
        elif j["cmd"] == "GET_GUILD":
            # We currently only get here because of a "CHANNEL_CREATE" event. Stupidly long winded way around
            if j["data"]:
                guild = j["data"]
            self.dump_channel_data()

            return
        elif j["cmd"] == "GET_CHANNELS":
            if j['evt'] == 'ERROR':
                log.error('%s', j['data']['message'])
                return
            self.guilds[j['nonce']]["channels"] = j["data"]["channels"]
            for channel in j["data"]["channels"]:
                channel['guild_id'] = j['nonce']
                channel['guild_name'] = self.guilds[j['nonce']]["name"]
                self.channels[channel["id"]] = channel
                if channel["type"] == 2:
                    self.req_channel_details(channel["id"])
            self.dump_channel_data()
            return
        elif j["cmd"] == "SUBSCRIBE":
            # Only log errors
            if j['evt']:
                log.warning(j)
            return
        elif j["cmd"] == "UNSUBSCRIBE":
            return
        elif j["cmd"] == "GET_SELECTED_VOICE_CHANNEL":
            if 'data' in j and j['data'] and 'id' in j['data']:
                self.set_channel(j['data']['id'], j['data']['guild_id'])
                self.discover.voice_overlay.set_channel_title(
                    j["data"]["name"])
                if self.current_guild in self.guilds and 'icon_url' in self.guilds[self.current_guild]:
                    self.discover.voice_overlay.set_channel_icon(
                        self.guilds[self.current_guild]['icon_url'])
                else:
                    self.discover.voice_overlay.set_channel_icon(None)
                self.list_altered = True
                self.in_room = []
                for u in j['data']['voice_states']:
                    thisuser = u["user"]
                    nick = u["nick"]
                    thisuser["nick"] = nick
                    mute = (u["voice_state"]["mute"] or
                            u["voice_state"]["self_mute"] or
                            u["voice_state"]["suppress"])
                    deaf = u["voice_state"]["deaf"] or u["voice_state"]["self_deaf"]
                    thisuser["mute"] = mute
                    thisuser["deaf"] = deaf
                    self.update_user(thisuser)
                    self.set_in_room(thisuser["id"], True)
            return
        elif j["cmd"] == "GET_CHANNEL":
            if j["evt"] == "ERROR":
                log.info(
                    "Could not get room")
                return
            if j["nonce"] == "new":
                self.req_channels(j["data"]["guild_id"])
            if j["data"]["type"] == 0:  # Text channel
                if self.current_text == j["data"]["id"]:
                    self.text = []
                    for message in j["data"]["messages"]:
                        self.add_text(message)

            return
        elif j["cmd"] == "SELECT_VOICE_CHANNEL":
            return
        elif j["cmd"] == "SET_VOICE_SETTINGS":
            return
        elif j["cmd"] == "GET_VOICE_SETTINGS":
            return
        log.warning(j)

    def dump_channel_data(self):
        with open(self.discover.channel_file, 'w') as f:
            f.write(json.dumps(
                {'channels': self.channels, 'guild': self.guilds}))

    def on_connected(self):
        """
        Called when connection is finalised
        """
        self.sub_server()
        self.find_user()
        if self.current_text:
            self.start_listening_text(self.current_text)

    def on_error(self, error):
        """
        Called when an error has occured
        """
        log.error("ERROR : %s", error)

    def on_close(self):
        """
        Called when connection is closed
        """
        log.warning("Connection closed")
        if self.socket_watch:
            GLib.source_remove(self.socket_watch)
            self.socket_watch = None
        self.websocket = None
        self.update_overlays_from_data()
        self.current_voice = "0"
        self.schedule_reconnect()

    def req_auth(self):
        """
        Request authentication token
        """
        cmd = {
            "cmd": "AUTHENTICATE",
            "args": {
                "access_token": self.access_token
            },
            "nonce": "deadbeef"
        }
        self.websocket.send(json.dumps(cmd))

    def req_guild(self, guild_id, nonce):
        """
        Request info on one guild
        """
        cmd = {
            "cmd": "GET_GUILD",
            "args": {"guild_id": guild_id},
            "nonce": nonce
        }
        self.websocket.send(json.dumps(cmd))

    def req_guilds(self):
        """
        Request all guilds information for logged in user
        """
        cmd = {
            "cmd": "GET_GUILDS",
            "args": {},
            "nonce": "deadbeef"
        }
        self.websocket.send(json.dumps(cmd))

    def req_channels(self, guild):
        """
        Request all channels information for given guild.
        Don't perform now but pass off to rate-limiter
        """

        if guild in self.guilds:
            self.rate_limited_channels.append(guild)
        else:
            log.warning(f"Didn't find guild with id {guild}")

    def req_channel_details(self, channel, nonce=None):
        """message
        Request information about a specific channel
        """
        if not self.websocket:
            return
        if not nonce:
            nonce = channel
        cmd = {
            "cmd": "GET_CHANNEL",
            "args": {
                "channel_id": channel
            },
            "nonce": nonce
        }
        self.websocket.send(json.dumps(cmd))

    def find_user(self):
        """
        Find the user
        """

        cmd = {
            "cmd": "GET_SELECTED_VOICE_CHANNEL",
            "args": {

            },
            "nonce": "test"
        }
        self.websocket.send(json.dumps(cmd))

    def sub_raw(self, event, args, nonce):
        """
        Subscribe to event helper function
        """
        cmd = {
            "cmd": "SUBSCRIBE",
            "args": args,
            "evt": event,
            "nonce": nonce
        }
        self.websocket.send(json.dumps(cmd))

    def unsub_raw(self, event, args, nonce):
        """
        Subscribe to event helper function
        """
        cmd = {
            "cmd": "UNSUBSCRIBE",
            "args": args,
            "evt": event,
            "nonce": nonce
        }
        self.websocket.send(json.dumps(cmd))

    def sub_server(self):
        """
        Subscribe to helpful events that report connectivity issues &
        when the user has intentionally changed channel

        Unfortunatly no event has been found to alert to being forcibly moved
        or that reports the users current location
        """
        self.sub_raw("VOICE_CHANNEL_SELECT", {}, "VOICE_CHANNEL_SELECT")
        self.sub_raw("VOICE_SETTINGS_UPDATE", {}, "VOICE_SETTINGS_UPDATE")
        self.sub_raw("VOICE_CONNECTION_STATUS", {}, "VOICE_CONNECTION_STATUS")
        self.sub_raw("GUILD_CREATE", {}, "GUILD_CREATE")
        self.sub_raw("CHANNEL_CREATE", {}, "CHANNEL_CREATE")
        self.sub_raw("NOTIFICATION_CREATE", {}, "NOTIFICATION_CREATE")

    def sub_channel(self, event, channel):
        """
        Subscribe to event on channel
        """
        self.sub_raw(event, {"channel_id": channel}, channel)

    def unsub_channel(self, event, channel):
        """
        Subscribe to event on channel
        """
        self.unsub_raw(event, {"channel_id": channel}, channel)

    def sub_text_channel(self, channel):
        """
        Subscribe to text-based events.
        """
        self.sub_channel("MESSAGE_CREATE", channel)
        self.sub_channel("MESSAGE_UPDATE", channel)
        self.sub_channel("MESSAGE_DELETE", channel)

    def unsub_text_channel(self, channel):
        """
        Unsubscribe to text-based events.
        """
        self.unsub_channel("MESSAGE_CREATE", channel)
        self.unsub_channel("MESSAGE_UPDATE", channel)
        self.unsub_channel("MESSAGE_DELETE", channel)

    def sub_voice_channel(self, channel):
        """
        Subscribe to voice-based events
        """
        self.sub_channel("VOICE_STATE_CREATE", channel)
        self.sub_channel("VOICE_STATE_UPDATE", channel)
        self.sub_channel("VOICE_STATE_DELETE", channel)
        self.sub_channel("SPEAKING_START", channel)
        self.sub_channel("SPEAKING_STOP", channel)

    def unsub_voice_channel(self, channel):
        """
        Remove subscription to voice-based events
        """
        self.unsub_channel("VOICE_STATE_CREATE", channel)
        self.unsub_channel("VOICE_STATE_UPDATE", channel)
        self.unsub_channel("VOICE_STATE_DELETE", channel)
        self.unsub_channel("SPEAKING_START", channel)
        self.unsub_channel("SPEAKING_STOP", channel)

    def get_voice_settings(self):
        """
        Request a recent version of voice settings
        """
        cmd = {
            "cmd": "GET_VOICE_SETTINGS",
            "args": {},
            "nonce": "deadbeef"
        }
        if self.websocket:
            self.websocket.send(json.dumps(cmd))

    def set_mute(self, muted):
        cmd = {
            "cmd": "SET_VOICE_SETTINGS",
            "args": {"mute": muted},
            "nonce": "deadbeef"
        }
        if self.websocket:
            self.websocket.send(json.dumps(cmd))
        return False

    def set_deaf(self, deaf):
        cmd = {
            "cmd": "SET_VOICE_SETTINGS",
            "args": {"deaf": deaf},
            "nonce": "deadbeef"
        }
        if self.websocket:
            self.websocket.send(json.dumps(cmd))
        return False

    def change_voice_room(self, id):
        """
        Switch to another voice room
        """
        cmd = {
            "cmd": "SELECT_VOICE_CHANNEL",
            "args": {
                "channel_id": id,
                "force": True
            },
            "nonce": "deadbeef"
        }
        if self.websocket:
            self.websocket.send(json.dumps(cmd))

    def change_text_room(self, id):
        """
        Switch to another text room
        """
        cmd = {
            "cmd": "SELECT_TEXT_CHANNEL",
            "args": {
                "channel_id": id
            },
            "nonce": "deadbeef"
        }
        if self.websocket:
            self.websocket.send(json.dumps(cmd))

    def update_overlays_from_data(self):
        if self.websocket == None:
            self.discover.voice_overlay.set_blank()
            if self.discover.text_overlay:
                self.discover.text_overlay.set_blank()
            if self.discover.notification_overlay:
                self.discover.notification_overlay.set_blank()
            return
        newlist = []
        for userid in self.in_room:
            newlist.append(self.userlist[userid])
        self.discover.voice_overlay.set_user_list(newlist, self.list_altered)
        self.list_altered = False
        # Update text list
        if self.discover.text_overlay.popup_style:
            self.text_altered = True
        if self.text_altered:
            self.discover.text_overlay.set_text_list(
                self.text, self.text_altered)
            self.text_altered = False

        if self.authed and len(self.rate_limited_channels) > 0:
            now = time.time()
            if self.last_rate_limit_send < now - 60:
                guild = self.rate_limited_channels.pop()

                cmd = {
                    "cmd": "GET_CHANNELS",
                    "args": {
                        "guild_id": guild
                    },
                    "nonce": guild
                }
                self.websocket.send(json.dumps(cmd))
                self.last_rate_limit_send = now

    def start_listening_text(self, channel):
        """
        Subscribe to text events on channel, or remember the channel for when we've connected

        Helper function to avoid race conditions of reading config vs connecting to websocket
        """
        if self.websocket:
            if self.current_text != "0":
                self.unsub_text_channel(self.current_text)
            if channel != "0":
                self.sub_text_channel(channel)
                self.req_channel_details(channel)
        self.current_text = channel

    def request_text_rooms_for_guild(self, guild_id):
        """
        Request a correctly ordered list of text channels.

        This will be mixed in with 'None' in the list where a voice channel is
        """
        if (guild_id == 0):
            return
        self.rate_limited_channels.append(guild_id)

    def schedule_reconnect(self):
        if self.reconnect_cb == None:
            log.info("Scheduled a reconnect")
            self.reconnect_cb = GLib.timeout_add_seconds(60, self.connect)
        else:
            log.error("Reconnect already scheduled")

    def connect(self):
        """
        Attempt to connect to websocket

        Should not throw simply for being unable to connect, only for more serious issues
        """
        log.info("Connecting...")
        if self.websocket:
            log.warn("Already connected?")
            return
        if self.reconnect_cb:
            GLib.source_remove(self.reconnect_cb)
            self.reconnect_cb = None
        try:
            self.websocket = websocket.create_connection(
                "ws://127.0.0.1:6463/?v=1&client_id=%s" % (self.oauth_token),
                origin="http://localhost:3000",
                timeout=0.1
            )
            if self.socket_watch:
                GLib.source_remove(self.socket_watch)
            self.socket_watch = GLib.io_add_watch(
                self.websocket.sock, GLib.PRIORITY_DEFAULT_IDLE, GLib.IOCondition.HUP | GLib.IOCondition.IN | GLib.IOCondition.ERR, self.socket_glib)
        except ConnectionError as error:
            self.schedule_reconnect()

    def socket_glib(self, fd, condition):
        if condition == GLib.IO_IN and self.websocket:
            recv, _w, _e = select.select((self.websocket.sock,), (), (), 0)
            while recv:
                try:
                    # Receive & send to on_message
                    msg = self.websocket.recv()
                    self.on_message(msg)
                    if not self.websocket:
                        # Connection was closed in the meantime
                        break
                    recv, _w, _e = select.select(
                        (self.websocket.sock,), (), (), 0)
                except (websocket.WebSocketConnectionClosedException, json.decoder.JSONDecodeError):
                    self.on_close()
                    break
            self.update_overlays_from_data()
        else:
            self.update_overlays_from_data()
            return False
        return True



#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""A class to assist with reading pulseaudio changes"""
import os
import logging
import signal
import pulsectl_asyncio
import pulsectl
from contextlib import suppress
import asyncio
from threading import Thread, Event

log = logging.getLogger(__name__)


class DiscoverAudioAssist:
    def __init__(self, discover):

        self.thread = None
        self.enabled = False
        self.source = None  # String containing the name of the PA/PW microphone or other input
        self.sink = None  # String containing the name of the PA/PW output

        self.discover = discover

        # Keep last known state (or None) so that we don't repeatedly send messages for every little PA/PW signal
        self.last_set_mute = None
        self.last_set_deaf = None

    def set_enabled(self, enabled):
        self.enabled = enabled
        if enabled:
            self.start()

    def set_devices(self, sink, source):
        # Changed devices from client
        self.source = source
        self.sink = sink

    def start(self):
        if not self.enabled:
            return
        if not self.thread:
            self.thread = Thread(target=self.thread_loop)
            self.thread.start()

    def thread_loop(self):
        # Start an asyncio specific thread. Not the prettiest but I'm not rewriting from ground up for one feature
        log.info("Staring Audio subsystem assistance")
        loop = asyncio.new_event_loop()
        loop.run_until_complete(self.pulse_loop())
        log.info("Stopped Audio subsystem assistance")

    async def listen(self):
        # Async to connect to pulse and listen for events
        try:
            async with pulsectl_asyncio.PulseAsync('Discover-Monitor') as pulse:
                await self.get_device_details(pulse)
                async for event in pulse.subscribe_events('all'):
                    await self.print_events(pulse, event)
        except (pulsectl.pulsectl.PulseDisconnected):
            log.info("Pulse has gone away")
        except (pulsectl.pulsectl.PulseError):
            log.info("Pulse error")

    async def pulse_loop(self):
        # Prep before connecting to pulse
        loop = asyncio.get_event_loop()
        listen_task = asyncio.create_task(self.listen())
        with suppress(asyncio.CancelledError):
            await listen_task

    async def get_device_details(self, pulse):
        # Decant information about our chosen devices
        # Feed this back to client to change deaf/mute state
        mute = None
        deaf = None
        for sink in await pulse.sink_list():
            if sink.description == self.sink:
                if sink.mute == 1 or sink.volume.values[0] == 0.0:
                    deaf = True
                elif sink.mute == 0:
                    deaf = False

        if deaf != self.last_set_deaf:
            self.last_set_deaf = deaf
            self.discover.set_deaf_async(deaf)
            self.last_set_mute = None
            # At this point mute is undefined state

        for source in await pulse.source_list():
            if source.description == self.source:
                if source.mute == 1 or source.volume.values[0] == 0.0:
                    mute = True
                elif sink.mute == 0:
                    mute = False

        if mute != self.last_set_mute:
            self.last_set_mute = mute
            self.discover.set_mute_async(mute)

    async def print_events(self, pulse, ev):
        if not self.enabled:
            return
        # Sink and Source events are fired for changes to output and ints
        # Server is fired when default sink or source changes.
        match ev.facility:
            case 'sink':
                await self.get_device_details(pulse)

            case 'source':
                await self.get_device_details(pulse)

            case 'server':
                await self.get_device_details(pulse)

            case 'source_output':
                pass

            case 'sink_input':
                pass

            case 'client':
                pass

            case _:
                # If we need to find more events, this here will do it
                # log.info('Pulse event: %s' % ev)
                pass
